#!/usr/bin/env python3

# Libervia Web: Libervia web frontend
# Copyright (C) 2013-2021 Jérôme Poisson <goffi@goffi.org>
# Copyright (C) 2013-2016 Adrien Cossa <souliane@mailoo.org>
# Copyright (C) 2013  Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twisted.internet import defer

if defer.Deferred.debug:
    # if we are in debug mode, we want to use ipdb instead of pdb
    try:
        import ipdb
        import pdb

        pdb.set_trace = ipdb.set_trace
        pdb.post_mortem = ipdb.post_mortem
    except ImportError:
        pass

import re
import os.path
import libervia
import sat

from libervia.server.constants import Const as C

from sat.core.i18n import _
from sat.tools import config

from zope.interface import implementer

from twisted.python import usage
from twisted.plugin import IPlugin
from twisted.application.service import IServiceMaker
import configparser


RE_VER_POST = re.compile(r"\.post[0-9]+")

if RE_VER_POST.sub("", libervia.__version__) != RE_VER_POST.sub("", sat.__version__):
    import sys

    sys.stderr.write(
        """sat module version ({sat_version}) and {current_app} version ({current_version}) mismatch

sat module is located at {sat_path}
libervia module is located at {libervia_path}

Please be sure to have the same version running
""".format(
            sat_version=sat.__version__,
            current_app=C.APP_NAME,
            current_version=libervia.__version__,
            sat_path=os.path.dirname(sat.__file__),
            libervia_path=os.path.dirname(libervia.__file__),
        )
    )
    sys.stderr.flush()
    # we call os._exit to avoid help to be printed by twisted
    import os

    os._exit(1)


def coerceConnectionType(value):  # called from Libervia.OPT_PARAMETERS
    assert isinstance(value, str)
    allowed_values = ("http", "https", "both")
    if value not in allowed_values:
        raise ValueError(
            "%(given)s not in %(expected)s"
            % {"given": value, "expected": str(allowed_values)}
        )
    return value


def coerceBool(value):
    return C.bool(value)


def coerceUnicode(value):
    assert isinstance(value, str)
    # XXX: we use this method to check which value to convert to Unicode
    #      but we don't do the conversion here as Twisted expect str
    return value


DATA_DIR_DEFAULT = ''
# prefix used for environment variables
ENV_PREFIX = "LIBERVIA_"
# options which are in sat.conf and on command line,
# see https://twistedmatrix.com/documents/current/api/twisted.python.usage.Options.html
OPT_PARAMETERS_BOTH = [['connection_type', 't', 'https', _("'http', 'https' or 'both' "
                        "(to launch both servers)."),
                        coerceConnectionType],
                       ['port', 'p', 8080,
                        _('The port number to listen HTTP on.'), int],
                       ['port_https', 's', 8443,
                        _('The port number to listen HTTPS on.'), int],
                       ['port_https_ext', 'e', 0, _('The external port number used for '
                        'HTTPS (0 means port_https value).'), int],
                       ['tls_private_key', '', '', _('TLS certificate private key (PEM '
                        'format)'), coerceUnicode],
                       ['tls_certificate', 'c', 'libervia.pem', _('TLS public '
                        'certificate or private key and public certificate combined '
                        '(PEM format)'), coerceUnicode],
                       ['tls_chain', '', '', _('TLS certificate intermediate chain (PEM '
                        'format)'), coerceUnicode],
                       ['redirect_to_https', 'r', True, _('Automatically redirect from '
                        'HTTP to HTTPS.'), coerceBool],
                       ['security_warning', 'w', True, _('Warn user that he is about to '
                        'connect on HTTP.'), coerceBool],
                       ['passphrase', 'k', '', (_("Passphrase for the SàT profile "
                        "named '%s'") % C.SERVICE_PROFILE),
                        coerceUnicode],
                       ['allow_registration', '', True, _('Allow user to register new '
                        'account'), coerceBool],
                       ['base_url_ext', '', '',
                        _('The external URL to use as base URL'),
                        coerceUnicode],
                       ['bridge-retries', '', 10,
                        _('Number of tries to connect to bridge before giving up'), int],
                      ]

# Options which are in sat.conf only
OPT_PARAMETERS_CFG = [
    ["empty_password_allowed_warning_dangerous_list", None, "", None],
    ["vhosts_dict", None, {}, None],
    ["url_redirections_dict", None, {}, None],
    ["menu_json", None, {'': C.DEFAULT_MENU}, None],
    ["menu_extra_json", None, {}, None],
    ["lists_directory_json", None, None, None],
    ["mr_handlers_json", None, None, None],
]

# Flags are in command line only
OPT_FLAGS = [
    ['build-only', 'B', _("Only build website, don't run the server")],
    ['dev-mode', 'D', _('Developer mode, automatically reload modified pages')],
]



def initialise(options):
    """Method to initialise global modules"""
    # XXX: We need to configure logs before any log method is used,
    #      so here is the best place.
    from sat.core import log_config

    log_config.satConfigure(C.LOG_BACKEND_TWISTED, C, backend_data=options)


class Options(usage.Options):
    # optArgs is not really useful in our case, we need more than a flag
    optParameters = OPT_PARAMETERS_BOTH
    optFlags = OPT_FLAGS

    def __init__(self):
        """Read SàT configuration file in order to overwrite the hard-coded default values

        Priority for the usage of the values is (from lowest to highest):
            - hard-coded default values
            - values from SàT configuration files
            - values passed on the command line
        """
        # If we do it the reading later: after the command line options have been parsed,
        # there's no good way to know
        # if the  options values are the hard-coded ones or if they have been passed
        # on the command line.

        # FIXME: must be refactored + code can be factorised with backend
        config_parser = config.parseMainConf()
        self.handleDeprecated(config_parser)
        for param in self.optParameters + OPT_PARAMETERS_CFG:
            name = param[0]
            env_name = f"{ENV_PREFIX}{name.upper()}"
            try:
                value = os.getenv(env_name)
                if value is None:
                    value = config.getConfig(
                        config_parser, C.CONFIG_SECTION, name, Exception)
                try:
                    param[2] = param[4](value)
                except IndexError:  # the coerce method is optional
                    param[2] = value
            except (configparser.NoSectionError, configparser.NoOptionError):
                pass
        usage.Options.__init__(self)
        for opt_data in OPT_PARAMETERS_CFG:
            self[opt_data[0]] = opt_data[2]

    def handleDeprecated(self, config_parser):
        """display warning and/or change option when a deprecated option if found

        param config_parser(ConfigParser): read ConfigParser instance for sat.conf
        """
        replacements = (("ssl_certificate", "tls_certificate"),)
        for old, new in replacements:
            try:
                value = config.getConfig(config_parser, C.CONFIG_SECTION, old, Exception)
            except (configparser.NoSectionError, configparser.NoOptionError):
                pass
            else:
                print(("\n/!\\ Use of {old} is deprecated, please use {new} instead\n"
                      .format(old=old, new=new)))
                config_parser.set(C.CONFIG_SECTION, new, value)


@implementer(IServiceMaker, IPlugin)
class LiberviaMaker(object):

    tapname = C.APP_NAME_FILE
    description = _("The web frontend of Libervia")
    options = Options

    def makeService(self, options):
        from twisted.internet import gireactor
        gireactor.install()
        for opt in OPT_PARAMETERS_BOTH:
            # FIXME: that's a ugly way to get unicode in Libervia
            #        from command line or sat.conf
            #        we should move to argparse and handle options this properly
            try:
                coerce_cb = opt[4]
            except IndexError:
                continue
            if coerce_cb == coerceUnicode:
                if not isinstance(options[opt[0]], str):
                    print(f"FIXME: {opt[0]} is not unicode")
                    options[opt[0]] = options[opt[0]].decode("utf-8")
        initialise(options.parent)
        from libervia.server import server

        return server.Libervia(options)


# affectation to some variable is necessary for twisted introspection to work
serviceMaker = LiberviaMaker()



#!/usr/bin/env python3

from libervia.server.constants import Const as C
from sat.tools.common import uri
import time

name = "blog_feed_atom"
access = C.PAGES_ACCESS_PUBLIC
template = "blog/atom.xml"


async def prepare_render(self, request):
    request.setHeader("Content-Type", "application/atom+xml; charset=utf-8")
    data = self.getRData(request)
    service, node = data["service"], data.get("node")
    self.checkCache(
        request, C.CACHE_PUBSUB, service=service, node=node, short="microblog"
    )
    data["show_comments"] = False
    template_data = request.template_data
    blog_page = self.getPageByName("blog_view")
    await blog_page.prepare_render(self, request)
    items = data["blog_items"]['items']

    template_data["request_uri"] = self.host.getExtBaseURL(
        request, request.path.decode("utf-8")
    )
    template_data["xmpp_uri"] = uri.buildXMPPUri(
        "pubsub", subtype="microblog", path=service.full(), node=node
    )
    blog_view = self.getPageByName("blog_view")
    template_data["http_uri"] = self.host.getExtBaseURL(
        request, blog_view.getURL(service.full(), node)
    )
    if items:
        template_data["updated"] = items[0]['updated']
    else:
        template_data["updated"] = time.time()



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from twisted.internet import defer
from sat.core.i18n import _
from sat.core.log import getLogger

log = getLogger(__name__)

name = "photos"
access = C.PAGES_ACCESS_PROFILE
template = "photo/discover.html"


@defer.inlineCallbacks
def prepare_render(self, request):
    profile = self.getProfile(request)
    template_data = request.template_data
    namespace = self.host.ns_map["fis"]
    if profile is not None:
        try:
            interests = yield self.host.bridgeCall(
                "interestsList", "", "", namespace, profile)
        except Exception:
            log.warning(_("Can't get interests list for {profile}").format(
                profile=profile))
        else:
            # we only want photo albums
            filtered_interests = []
            for interest in interests:
                if interest.get('subtype') != 'photos':
                    continue
                path = interest.get('path', '')
                path_args = [p for p in path.split('/') if p]
                interest["url"] = self.getSubPageURL(
                    request,
                    "photos_album",
                    interest['service'],
                    *path_args
                )
                filtered_interests.append(interest)

            template_data['interests'] = filtered_interests

        template_data["url_photos_new"] = self.getSubPageURL(request, "photos_new")


@defer.inlineCallbacks
def on_data_post(self, request):
    jid_ = self.getPostedData(request, "jid")
    url = self.getPageByName("photos_album").getURL(jid_)
    self.HTTPRedirect(request, url)



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from twisted.internet import defer
from twisted.words.protocols.jabber import jid

"""page used to target a user profile, e.g. for public blog"""

name = "user"
access = C.PAGES_ACCESS_PUBLIC  # can be a callable
template = "blog/articles.html"
url_cache = True


@defer.inlineCallbacks
def parse_url(self, request):
    try:
        prof_requested = self.nextPath(request)
    except IndexError:
        self.pageError(request)

    data = self.getRData(request)

    target_profile = yield self.host.bridgeCall("profileNameGet", prof_requested)
    request.template_data["target_profile"] = target_profile
    target_jid = yield self.host.bridgeCall(
        "asyncGetParamA", "JabberID", "Connection", "value", profile_key=target_profile
    )
    target_jid = jid.JID(target_jid)
    data["service"] = target_jid

    # if URL is parsed here, we'll have atom.xml available and we need to
    # add the link to the page
    atom_url = self.getSubPageURL(request, 'user_blog_feed_atom')
    request.template_data['atom_url'] = atom_url
    request.template_data.setdefault('links', []).append({
        "href": atom_url,
        "type": "application/atom+xml",
        "rel": "alternate",
        "title": "{target_profile}'s blog".format(target_profile=target_profile)})

def add_breadcrumb(self, request, breadcrumbs):
    # we don't want a breadcrumb here
    pass


@defer.inlineCallbacks
def prepare_render(self, request):
    data = self.getRData(request)
    self.checkCache(
        request, C.CACHE_PUBSUB, service=data["service"], node=None, short="microblog"
    )
    self.pageRedirect("blog_view", request)

def on_data_post(self, request):
    return self.getPageByName("blog_view").on_data_post(self, request)



from browser import document, window, bind, html, DOMNode, aio
from javascript import JSON
from bridge import Bridge
from aio_bridge import Bridge as AIOBridge
from template import Template
import dialog
from slideshow import SlideShow
from invitation import InvitationManager
import alt_media_player
# we use tmp_aio because `blob` is not handled in Brython's aio
import tmp_aio
import loading


cache_path = window.cache_path
files_service = window.files_service
files_path = window.files_path
try:
    affiliations = window.affiliations.to_dict()
except AttributeError:
    pass
bridge = Bridge()
aio_bridge = AIOBridge()

alt_media_player.install_if_needed()

photo_tpl = Template('photo/item.html')
player_tpl = Template('components/media_player.html')

# file upload

def on_progress(ev, photo_elt):
    if ev.lengthComputable:
        percent = int(ev.loaded/ev.total*100)
        update_progress(photo_elt, percent)


def on_load(file_, photo_elt):
    update_progress(photo_elt, 100)
    photo_elt.classList.add("progress_finished")
    photo_elt.classList.remove("progress_started")
    photo_elt.select_one('.action_delete').bind("click", on_delete)
    print(f"file {file_.name} uploaded correctly")


def on_error(failure, file_, photo_elt):
    dialog.notification.show(
        f"can't upload {file_.name}: {failure}",
        level="error"
    )


def update_progress(photo_elt, new_value):
    progress_elt = photo_elt.select_one("progress")
    progress_elt.value = new_value
    progress_elt.text = f"{new_value}%"


def on_slot_cb(file_, upload_slot, photo_elt):
    put_url, get_url, headers = upload_slot
    xhr = window.XMLHttpRequest.new()
    xhr.open("PUT", put_url, True)
    xhr.upload.bind('progress', lambda ev: on_progress(ev, photo_elt))
    xhr.upload.bind('load', lambda ev: on_load(file_, photo_elt))
    xhr.upload.bind('error', lambda ev: on_error(xhr.response, file_, photo_elt))
    xhr.setRequestHeader('Xmpp-File-Path', window.encodeURIComponent(files_path))
    xhr.setRequestHeader('Xmpp-File-No-Http', "true")
    xhr.send(file_)


def on_slot_eb(file_, failure, photo_elt):
    dialog.notification.show(
        f"Can't get upload slot: {failure['message']}",
        level="error"
    )
    photo_elt.remove()


def upload_files(files):
    print(f"uploading {len(files)} files")
    album_items = document['album_items']
    for file_ in files:
        url = window.URL.createObjectURL(file_)
        photo_elt = photo_tpl.get_elt({
            "file": {
                "name": file_.name,
                # we don't want to open the file on click, it's not yet the
                # uploaded URL
                "url": url,
                # we have no thumb yet, so we use the whole image
                # TODO: reduce image for preview
                "thumb_url": url,
            },
            "uploading": True,
        })
        photo_elt.classList.add("progress_started")
        album_items <= photo_elt

        bridge.fileHTTPUploadGetSlot(
            file_.name,
            file_.size,
            file_.type or '',
            files_service,
            callback=lambda upload_slot, file_=file_, photo_elt=photo_elt:
                on_slot_cb(file_, upload_slot, photo_elt),
            errback=lambda failure, file_=file_, photo_elt=photo_elt:
                on_slot_eb(file_, failure, photo_elt),
        )


@bind("#file_drop", "drop")
def on_file_select(evt):
    evt.stopPropagation()
    evt.preventDefault()
    files = evt.dataTransfer.files
    upload_files(files)


@bind("#file_drop", "dragover")
def on_drag_over(evt):
    evt.stopPropagation()
    evt.preventDefault()
    evt.dataTransfer.dropEffect = 'copy'


@bind("#file_input", "change")
def on_file_input_change(evt):
    files = evt.currentTarget.files
    upload_files(files)

# delete

def file_delete_cb(item_elt, item):
    item_elt.classList.add("state_deleted")
    item_elt.bind("transitionend", lambda evt: item_elt.remove())
    print(f"deleted {item['name']}")


def file_delete_eb(failure, item_elt, item):
    dialog.notification.show(
        f"error while deleting {item['name']}: failure",
        level="error"
    )


def delete_ok(evt, notif_elt, item_elt, item):
    file_path = f"{files_path.rstrip('/')}/{item['name']}"
    bridge.fileSharingDelete(
        files_service,
        file_path,
        "",
        callback=lambda : file_delete_cb(item_elt, item),
        errback=lambda failure: file_delete_eb(failure, item_elt, item),
    )


def delete_cancel(evt, notif_elt, item_elt, item):
    notif_elt.remove()
    item_elt.classList.remove("selected_for_deletion")


def on_delete(evt):
    evt.stopPropagation()
    target = evt.currentTarget
    item_elt = DOMNode(target.closest('.item'))
    item_elt.classList.add("selected_for_deletion")
    item = JSON.parse(item_elt.dataset.item)
    dialog.Confirm(
        f"{item['name']!r} will be deleted, are you sure?",
        ok_label="delete",
        ok_color="danger",
    ).show(
        ok_cb=lambda evt, notif_elt: delete_ok(evt, notif_elt, item_elt, item),
        cancel_cb=lambda evt, notif_elt: delete_cancel(evt, notif_elt, item_elt, item),
    )

# cover

async def cover_ok(evt, notif_elt, item_elt, item):
    # we first need to get a blob of the image
    img_elt = item_elt.select_one("img")
    # the simplest way is to download it
    r = await tmp_aio.ajax("GET", img_elt.src, "blob")
    if r.status != 200:
        dialog.notification.show(
            f"can't retrieve cover: {r.status}: {r.statusText}",
            level="error"
        )
        return
    img_blob = r.response
    # now we'll upload it via HTTP Upload, we need a slow
    img_name = img_elt.src.rsplit('/', 1)[-1]
    img_size = img_blob.size

    slot = await aio_bridge.fileHTTPUploadGetSlot(
        img_name,
        img_size,
        '',
        files_service
    )
    get_url, put_url, headers = slot
    # we have the slot, we can upload image
    r = await tmp_aio.ajax("PUT", put_url, "", img_blob)
    if r.status != 201:
        dialog.notification.show(
            f"can't upload cover: {r.status}: {r.statusText}",
            level="error"
        )
        return
    extra = {"thumb_url": get_url}
    album_name = files_path.rsplit('/', 1)[-1]
    await aio_bridge.interestsRegisterFileSharing(
        files_service,
        "photos",
        "",
        files_path,
        album_name,
        JSON.stringify(extra),
    )
    dialog.notification.show("Album cover has been changed")


def cover_cancel(evt, notif_elt, item_elt, item):
    notif_elt.remove()
    item_elt.classList.remove("selected_for_action")


def on_cover(evt):
    evt.stopPropagation()
    target = evt.currentTarget
    item_elt = DOMNode(target.closest('.item'))
    item_elt.classList.add("selected_for_action")
    item = JSON.parse(item_elt.dataset.item)
    dialog.Confirm(
        f"use {item['name']!r} for this album cover?",
        ok_label="use as cover",
    ).show(
        ok_cb=lambda evt, notif_elt: aio.run(cover_ok(evt, notif_elt, item_elt, item)),
        cancel_cb=lambda evt, notif_elt: cover_cancel(evt, notif_elt, item_elt, item),
    )


# slideshow

@bind(".photo_thumb_click", "click")
def photo_click(evt):
    evt.stopPropagation()
    evt.preventDefault()
    slideshow = SlideShow()
    target = evt.currentTarget
    clicked_item_elt = DOMNode(target.closest('.item'))

    slideshow.attach()
    for idx, item_elt in enumerate(document.select('.item')):
        item = JSON.parse(item_elt.dataset.item)
        try:
            biggest_thumb = item['extra']['thumbnails'][-1]
            thumb_url = f"{cache_path}{biggest_thumb['filename']}"
        except (KeyError, IndexError) as e:
            print(f"Can't get full screen thumbnail URL: {e}")
            thumb_url = None
        if item.get("mime_type", "")[:5] == "video":
            player = alt_media_player.MediaPlayer(
                [item['url']],
                poster = thumb_url,
                reduce_click_area =  True
            )
            elt = player.elt
            elt.classList.add("slide_video", "no_fullscreen")
            slideshow.add_slide(
                elt,
                item,
                options={
                    "flags": (alt_media_player.NO_PAGINATION, alt_media_player.NO_SCROLLBAR),
                    "exit_callback": player.reset,
                }
            )
        else:
            slideshow.add_slide(html.IMG(src=thumb_url or item['url'], Class="slide_img"), item)
        if item_elt == clicked_item_elt:
            slideshow.index = idx


for elt in document.select('.action_delete'):
    elt.bind("click", on_delete)
for elt in document.select('.action_cover'):
    elt.bind("click", on_cover)

# manage


@bind("#button_manage", "click")
def manage_click(evt):
    evt.stopPropagation()
    evt.preventDefault()
    manager = InvitationManager("photos", {"service": files_service, "path": files_path})
    manager.attach(affiliations=affiliations)


# hint
@bind("#hint .click_to_delete", "click")
def remove_hint(evt):
    document['hint'].remove()


loading.remove_loading_screen()



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from twisted.words.protocols.jabber import jid
from sat.core.log import getLogger

log = getLogger(__name__)
"""files handling pages"""

name = "files"
access = C.PAGES_ACCESS_PROFILE
template = "file/discover.html"


async def prepare_render(self, request):
    profile = self.getProfile(request)
    template_data = request.template_data
    namespace = self.host.ns_map["fis"]
    entities_services, entities_own, entities_roster = await self.host.bridgeCall(
        "discoFindByFeatures", [namespace], [], False, True, True, True, False, profile
    )
    tpl_service_entities = template_data["disco_service_entities"] = {}
    tpl_own_entities = template_data["disco_own_entities"] = {}
    tpl_roster_entities = template_data["disco_roster_entities"] = {}
    entities_url = template_data["entities_url"] = {}

    # we store identities in dict of dict using category and type as keys
    # this way it's easier to test category in the template
    for tpl_entities, entities_map in (
        (tpl_service_entities, entities_services),
        (tpl_own_entities, entities_own),
        (tpl_roster_entities, entities_roster),
    ):
        for entity_str, entity_ids in entities_map.items():
            entity_jid = jid.JID(entity_str)
            tpl_entities[entity_jid] = identities = {}
            for cat, type_, name in entity_ids:
                identities.setdefault(cat, {}).setdefault(type_, []).append(name)
            entities_url[entity_jid] = self.getPageByName("files_list").getURL(
                entity_jid.full()
            )


def on_data_post(self, request):
    jid_str = self.getPostedData(request, "jid")
    try:
        jid_ = jid.JID(jid_str)
    except RuntimeError:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    url = self.getPageByName("files_list").getURL(jid_.full())
    self.HTTPRedirect(request, url)



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from sat.core.i18n import _
from twisted.internet import defer
from sat.tools.common import template_xmlui
from sat.tools.common import data_format
from sat.core.log import getLogger

log = getLogger(__name__)

name = "list_edit"
access = C.PAGES_ACCESS_PROFILE
template = "list/edit.html"


def parse_url(self, request):
    self.getPathArgs(request, ["service", "node", "item_id"], service="jid", node="@")
    data = self.getRData(request)
    if data["item_id"] is None:
        log.warning(_("no list item id specified"))
        self.pageError(request, C.HTTP_BAD_REQUEST)

@defer.inlineCallbacks
def prepare_render(self, request):
    data = self.getRData(request)
    template_data = request.template_data
    service, node, item_id = (
        data.get("service", ""),
        data.get("node", ""),
        data["item_id"],
    )
    profile = self.getProfile(request)

    # we don't ignore "author" below to keep it when a list item is edited
    # by node owner/admin and "consistent publisher" is activated
    ignore = (
        "publisher",
        "author",
        "author_jid",
        "author_email",
        "created",
        "updated",
        "comments_uri",
    )
    list_raw = yield self.host.bridgeCall(
        "listGet",
        service.full() if service else "",
        node,
        C.NO_LIMIT,
        [item_id],
        "",
        {},
        profile,
    )
    list_items, metadata = data_format.deserialise(list_raw, type_check=list)
    list_item = [template_xmlui.create(self.host, x, ignore=ignore) for x in list_items][0]

    try:
        # small trick to get a one line text input instead of the big textarea
        list_item.widgets["labels"].type = "string"
        list_item.widgets["labels"].value = list_item.widgets["labels"].value.replace(
            "\n", ", "
        )
    except KeyError:
        pass

    # for now we don't have XHTML editor, so we'll go with a TextBox and a convertion
    # to a text friendly syntax using markdown
    wid = list_item.widgets['body']
    if wid.type == "xhtmlbox":
        wid.type = "textbox"
        wid.value =  yield self.host.bridgeCall(
            "syntaxConvert", wid.value, C.SYNTAX_XHTML, "markdown",
            False, profile)

    template_data["new_list_item_xmlui"] = list_item


@defer.inlineCallbacks
def on_data_post(self, request):
    data = self.getRData(request)
    service = data["service"]
    node = data["node"]
    item_id = data["item_id"]
    posted_data = self.getAllPostedData(request)
    if not posted_data["title"] or not posted_data["body"]:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    try:
        posted_data["labels"] = [l.strip() for l in posted_data["labels"][0].split(",")]
    except (KeyError, IndexError):
        pass
    profile = self.getProfile(request)

    # we convert back body to XHTML
    body = yield self.host.bridgeCall(
        "syntaxConvert", posted_data['body'][0], "markdown", C.SYNTAX_XHTML,
        False, profile)
    posted_data['body'] = ['<div xmlns="{ns}">{body}</div>'.format(ns=C.NS_XHTML,
                                                                     body=body)]

    extra = {'update': True}
    yield self.host.bridgeCall(
        "listSet", service.full(), node, posted_data, "", item_id,
        data_format.serialise(extra), profile
    )
    # we don't want to redirect to edit page on success, but to list overview
    data["post_redirect_page"] = (
        self.getPageByName("lists"),
        service.full(),
        node or "@",
    )



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from twisted.words.protocols.jabber import jid
from sat.tools.common.template import safe
from sat.tools.common import data_format
from sat.core.i18n import _, D_
from sat.core.log import getLogger
import time
import html
import math
import re

name = "event_admin"
label = D_("Event Administration")
access = C.PAGES_ACCESS_PROFILE
template = "event/admin.html"
log = getLogger(__name__)
REG_EMAIL_RE = re.compile(C.REG_EMAIL_RE, re.IGNORECASE)


def parse_url(self, request):
    self.getPathArgs(
        request,
        ("event_service", "event_node", "event_id"),
        min_args=2,
        event_service="@jid",
        event_id="",
    )


async def prepare_render(self, request):
    data = self.getRData(request)
    template_data = request.template_data

    ## Event ##

    event_service = template_data["event_service"] = data["event_service"]
    event_node = template_data["event_node"] = data["event_node"]
    event_id = template_data["event_id"] = data["event_id"]
    profile = self.getProfile(request)
    event_timestamp, event_data = await self.host.bridgeCall(
        "eventGet",
        event_service.userhost() if event_service else "",
        event_node,
        event_id,
        profile,
    )
    try:
        background_image = event_data.pop("background-image")
    except KeyError:
        pass
    else:
        template_data["dynamic_style"] = safe(
            """
            html {
                background-image: url("%s");
                background-size: 15em;
            }
            """
            % html.escape(background_image, True)
        )
    template_data["event"] = event_data
    invitees = await self.host.bridgeCall(
        "eventInviteesList",
        event_data["invitees_service"],
        event_data["invitees_node"],
        profile,
    )
    template_data["invitees"] = invitees
    invitees_guests = 0
    for invitee_data in invitees.values():
        if invitee_data.get("attend", "no") == "no":
            continue
        try:
            invitees_guests += int(invitee_data.get("guests", 0))
        except ValueError:
            log.warning(
                _("guests value is not valid: {invitee}").format(invitee=invitee_data)
            )
    template_data["invitees_guests"] = invitees_guests
    template_data["days_left"] = int(
        math.ceil((event_timestamp - time.time()) / (60 * 60 * 24))
    )

    ## Blog ##

    data["service"] = jid.JID(event_data["blog_service"])
    data["node"] = event_data["blog_node"]
    data["allow_commenting"] = "simple"

    # we now need blog items, using blog common page
    # this will fill the "items" template data
    blog_page = self.getPageByName("blog_view")
    await blog_page.prepare_render(self, request)


async def on_data_post(self, request):
    profile = self.getProfile(request)
    if not profile:
        log.error("got post data without profile")
        self.pageError(request, C.HTTP_INTERNAL_ERROR)
    type_ = self.getPostedData(request, "type")
    if type_ == "blog":
        service, node, title, body, lang = self.getPostedData(
            request, ("service", "node", "title", "body", "language")
        )

        if not body.strip():
            self.pageError(request, C.HTTP_BAD_REQUEST)
        data = {"content": body}
        if title:
            data["title"] = title
        if lang:
            data["language"] = lang
        try:
            comments = bool(self.getPostedData(request, "comments").strip())
        except KeyError:
            pass
        else:
            if comments:
                data["allow_comments"] = True

        try:
            await self.host.bridgeCall(
                "mbSend", service, node, data_format.serialise(data), profile)
        except Exception as e:
            if "forbidden" in str(e):
                self.pageError(request, C.HTTP_FORBIDDEN)
            else:
                raise e
    elif type_ == "event":
        service, node, event_id, jids, emails = self.getPostedData(
            request, ("service", "node", "event_id", "jids", "emails")
        )
        for invitee_jid_s in jids.split():
            try:
                invitee_jid = jid.JID(invitee_jid_s)
            except RuntimeError:
                log.warning(
                    _("this is not a valid jid: {jid}").format(jid=invitee_jid_s)
                )
                continue
            await self.host.bridgeCall(
                "eventInvite", invitee_jid.userhost(), service, node, event_id, profile
            )
        for email_addr in emails.split():
            if not REG_EMAIL_RE.match(email_addr):
                log.warning(
                    _("this is not a valid email address: {email}").format(
                        email=email_addr
                    )
                )
                continue
            await self.host.bridgeCall(
                "eventInviteByEmail",
                service,
                node,
                event_id,
                email_addr,
                {},
                "",
                "",
                "",
                "",
                "",
                "",
                profile,
            )

    else:
        log.warning(_("Unhandled data type: {}").format(type_))



from browser import document, window, aio
from aio_bridge import Bridge
import dialog

try:
    pubsub_service = window.pubsub_service
    pubsub_node = window.pubsub_node
    pubsub_item = window.pubsub_item
except AttributeError:
    can_delete = False
else:
    bridge = Bridge()
    can_delete = True


async def on_delete(evt):
    evt.stopPropagation()
    confirmed = await dialog.Confirm(
        "This item will be deleted, are you sure?",
        ok_label="delete",
        ok_color="danger",
    ).ashow()
    if confirmed:
        try:
            comments_service = window.comments_service
            comments_node = window.comments_node
        except AttributeError:
            pass
        else:
            print(f"deleting comment node at [{comments_service}] {comments_node!r}")
            try:
                await bridge.psNodeDelete(comments_service, comments_node)
            except Exception as e:
                dialog.notification.show(
                    f"Can't delete comment node: {e}",
                    level="error"
                )

        print(f"deleting list item {pubsub_item!r} at [{pubsub_service}] {pubsub_node!r}")
        try:
            await bridge.psItemRetract(pubsub_service, pubsub_node, pubsub_item, True)
        except Exception as e:
            dialog.notification.show(
                f"Can't delete list item: {e}",
                level="error"
            )
        else:
            # FIXME: Q&D way to get list view URL, need to have a proper method (would
            #   be nice to have a way to reference pages by name from browser)
            list_url = '/'.join(window.location.pathname.split('/')[:-1]).replace(
                'view_item', 'view')
            window.location.replace(list_url)


if can_delete:
    for elt in document.select('.action_delete'):
        elt.bind("click", lambda evt: aio.run(on_delete(evt)))



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from sat.core.log import getLogger

log = getLogger(__name__)


name = "merge-requests_new"
access = C.PAGES_ACCESS_PUBLIC
template = "merge-request/create.html"



#!/usr/bin/env python3

from libervia.server.constants import Const as C
from sat.core.log import getLogger
from sat.core.i18n import _
from sat.tools.common import uri as xmpp_uri

log = getLogger(__name__)
import json

"""forum handling pages"""

name = "forums"
access = C.PAGES_ACCESS_PUBLIC
template = "forum/overview.html"


def parse_url(self, request):
    self.getPathArgs(
        request,
        ["service", "node", "forum_key"],
        service="@jid",
        node="@",
        forum_key="",
    )


def add_breadcrumb(self, request, breadcrumbs):
    # we don't want breadcrumbs here as long as there is no forum discovery
    # because it will be the landing page for forums activity until then
    pass


def getLinks(self, forums):
    for forum in forums:
        try:
            uri = forum["uri"]
        except KeyError:
            pass
        else:
            uri = xmpp_uri.parseXMPPUri(uri)
            service = uri["path"]
            node = uri["node"]
            forum["http_url"] = self.getPageByName("forum_topics").getURL(service, node)
        if "sub-forums" in forum:
            getLinks(self, forum["sub-forums"])


async def prepare_render(self, request):
    data = self.getRData(request)
    template_data = request.template_data
    service, node, key = data["service"], data["node"], data["forum_key"]
    profile = self.getProfile(request) or C.SERVICE_PROFILE

    try:
        forums_raw = await self.host.bridgeCall(
            "forumsGet", service.full() if service else "", node, key, profile
        )
    except Exception as e:
        log.warning(_("Can't retrieve forums: {msg}").format(msg=e))
        forums = []
    else:
        forums = json.loads(forums_raw)
    getLinks(self, forums)

    template_data["forums"] = forums



#!/usr/bin/env python3

import json
from sat.core.i18n import _
from sat.core.log import getLogger
from sat_frontends.bridge.bridge_frontend import BridgeException
from libervia.server.constants import Const as C


log = getLogger(__name__)
"""access to restricted bridge"""

name = "bridge"
on_data_post = "continue"

# bridge method allowed when no profile is connected
NO_SESSION_ALLOWED = ("getContacts", "identitiesBaseGet", "identitiesGet")


def parse_url(self, request):
    self.getPathArgs(request, ["method_name"], min_args=1)


async def render(self, request):
    if request.method != b'POST':
        log.warning(f"Bad method used with _bridge endpoint: {request.method.decode()}")
        return self.pageError(request, C.HTTP_BAD_REQUEST)
    data = self.getRData(request)
    profile = self.getProfile(request)
    self.checkCSRF(request)
    method_name = data["method_name"]
    if profile is None:
        if method_name in NO_SESSION_ALLOWED:
            # this method is allowed, we use the service profile
            profile = C.SERVICE_PROFILE
        else:
            log.warning("_bridge endpoint accessed without authorisation")
            return self.pageError(request, C.HTTP_UNAUTHORIZED)
    method_data = json.load(request.content)
    try:
        bridge_method = getattr(self.host.restricted_bridge, method_name)
    except AttributeError:
        log.warning(_(
            "{profile!r} is trying to access a bridge method not implemented in "
            "RestrictedBridge: {method_name}").format(
                profile=profile, method_name=method_name))
        return self.pageError(request, C.HTTP_BAD_REQUEST)

    try:
        args, kwargs = method_data['args'], method_data['kwargs']
    except KeyError:
        log.warning(_(
            "{profile!r} has sent a badly formatted method call: {method_data}"
        ).format(profile=profile, method_data=method_data))
        return self.pageError(request, C.HTTP_BAD_REQUEST)

    if "profile" in kwargs or "profile_key" in kwargs:
        log.warning(_(
            '"profile" key should not be in method kwargs, hack attempt? '
            "profile={profile}, method_data={method_data}"
        ).format(profile=profile, method_data=method_data))
        return self.pageError(request, C.HTTP_BAD_REQUEST)

    try:
        ret = await bridge_method(*args, **kwargs, profile=profile)
    except BridgeException as e:
        request.setResponseCode(C.HTTP_PROXY_ERROR)
        ret = {
            "fullname": e.fullname,
            "message": e.message,
            "condition": e.condition,
            "module": e.module,
            "classname": e.classname,
        }
    return json.dumps(ret)



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from twisted.words.protocols.jabber import jid
from sat.core.log import getLogger

log = getLogger(__name__)


name = "merge-requests_disco"
access = C.PAGES_ACCESS_PUBLIC
template = "merge-request/discover.html"


def prepare_render(self, request):
    mr_handlers_config = self.host.options["mr_handlers_json"]
    if mr_handlers_config:
        handlers = request.template_data["mr_handlers"] = []
        try:
            for handler_data in mr_handlers_config:
                service = handler_data["service"]
                node = handler_data["node"]
                name = handler_data["name"]
                url = self.getPageByName("merge-requests").getURL(service, node)
                handlers.append({"name": name, "url": url})
        except KeyError as e:
            log.warning("Missing field in mr_handlers_json: {msg}".format(msg=e))
        except Exception as e:
            log.warning("Can't decode mr handlers: {msg}".format(msg=e))


def on_data_post(self, request):
    jid_str = self.getPostedData(request, "jid")
    try:
        jid_ = jid.JID(jid_str)
    except RuntimeError:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    # for now we just use default node
    url = self.getPageByName("merge-requests").getURL(jid_.full(), "@")
    self.HTTPRedirect(request, url)



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from sat.core.i18n import _, D_
from sat.core.log import getLogger
from sat.tools.common import data_format

log = getLogger(__name__)

name = "forum_view"
label = D_("View")
access = C.PAGES_ACCESS_PUBLIC
template = "forum/view.html"


def parse_url(self, request):
    self.getPathArgs(request, ["service", "node"], 2, service="jid")


async def prepare_render(self, request):
    data = self.getRData(request)
    data["show_comments"] = False
    blog_page = self.getPageByName("blog_view")
    request.args[b"before"] = [b""]
    request.args[b"reverse"] = [b"1"]
    await blog_page.prepare_render(self, request)
    request.template_data["login_url"] = self.getPageRedirectURL(request)


async def on_data_post(self, request):
    profile = self.getProfile(request)
    if profile is None:
        self.pageError(request, C.HTTP_FORBIDDEN)
    type_ = self.getPostedData(request, "type")
    if type_ == "comment":
        service, node, body = self.getPostedData(request, ("service", "node", "body"))

        if not body:
            self.pageError(request, C.HTTP_BAD_REQUEST)
        mb_data = {"content_rich": body}
        try:
            await self.host.bridgeCall(
                "mbSend", service, node, data_format.serialise(mb_data), profile)
        except Exception as e:
            if "forbidden" in str(e):
                self.pageError(request, 401)
            else:
                raise e
    else:
        log.warning(_("Unhandled data type: {}").format(type_))



from browser import window

"""Q&D module to do ajax requests with data types currently unsupported by Brython"""
# FIXME: remove this module when official aio module allows to work with blobs

window.eval("""
var _tmp_ajax = function(method, url, format, data){
    return new Promise(function(resolve, reject){
        var xhr = new XMLHttpRequest()
        xhr.open(method, url, true)
        xhr.responseType = format
        xhr.onreadystatechange = function(){
            if(this.readyState == 4){
                resolve(this)
            }
        }
        if(data){
            xhr.send(data)
        }else{
            xhr.send()
        }
    })
}
""")

ajax = window._tmp_ajax



#!/usr/bin/env python3

# Libervia: a Salut à Toi frontend
# Copyright (C) 2011-2021 Jérôme Poisson <goffi@goffi.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from twisted.web import proxy
from twisted.python.compat import urlquote
from twisted.internet import address
from sat.core.log import getLogger
from libervia.server.constants import Const as C

log = getLogger(__name__)



class SatProxyClient(proxy.ProxyClient):

    def handleHeader(self, key, value):
        if key.lower() == b"x-frame-options":
            value = b"sameorigin"
        elif key.lower() == b"content-security-policy":
            value = value.replace(b"frame-ancestors 'none'", b"frame-ancestors 'self'")

        super().handleHeader(key, value)


class SatProxyClientFactory(proxy.ProxyClientFactory):
    protocol = SatProxyClient


class SatReverseProxyResource(proxy.ReverseProxyResource):
    """Resource Proxy rewritting headers to allow embedding in iframe on same domain"""
    proxyClientFactoryClass = SatProxyClientFactory

    def getChild(self, path, request):
        return SatReverseProxyResource(
            self.host, self.port,
            self.path + b'/' + urlquote(path, safe=b"").encode('utf-8'),
            self.reactor
        )

    def render(self, request):
        # Forwarded and X-Forwarded-xxx headers can be set if we have behin an other proxy
        if ((not request.getHeader(C.H_FORWARDED)
             and not request.getHeader(C.H_X_FORWARDED_HOST))):
            forwarded_data = []
            addr = request.getClientAddress()
            if ((isinstance(addr, address.IPv4Address)
                 or isinstance(addr, address.IPv6Address))):
                request.requestHeaders.setRawHeaders(C.H_X_FORWARDED_FOR, [addr.host])
                forwarded_data.append(f"for={addr.host}")
            host = request.getHeader("host")
            if host is None:
                port = request.getHost().port
                hostname = request.getRequestHostname()
                host = hostname if port in (80, 443) else f"{hostname}:{port}"
            request.requestHeaders.setRawHeaders(C.H_X_FORWARDED_HOST, [host])
            forwarded_data.append(f"host={host}")
            proto = "https" if request.isSecure() else "http"
            request.requestHeaders.setRawHeaders(C.H_X_FORWARDED_PROTO, [proto])
            forwarded_data.append(f"proto={proto}")
            request.requestHeaders.setRawHeaders(
                C.H_FORWARDED, [";".join(forwarded_data)]
            )

        return super().render(request)



#!/usr/bin/env python3

# Libervia Web Frontend
# Copyright (C) 2011-2021  Jérôme Poisson (goffi@goffi.org)
# Copyright (C) 2013-2016 Adrien Cossa (souliane@mailoo.org)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup
import os

NAME = "libervia-web"
# NOTE: directory is still "libervia" for compatibility reason, should be changed for 0.9
DIR_NAME = "libervia"

install_requires = [
    "libervia-backend >=0.8.0b1, <0.9.0",
    "libervia-templates >=0.8.0b1, <0.9.0",
    'twisted[tls] >=20.3.0, <21.3.0',
    "zope.interface <5.5.0",
    'pyopenssl <21.0.0',
    "jinja2 >=2.9, <3.1",
    'shortuuid <1.1',
    "autobahn <21.4.0",
    "brython >=3.9.2, <3.10",
]
long_description = """\
Web frontend for Libervia (formerly Salut à Toi), a multi-frontends and multi-purposes XMPP client.
It features chat, blog, forums, events, tickets, merge requests, file sharing, photo albums, etc.
It is also a decentralized, XMPP based web framework.
"""

with open(os.path.join(DIR_NAME, "VERSION")) as v:
    VERSION = v.read().strip()
is_dev_version = VERSION.endswith("D")


def libervia_dev_version():
    """Use mercurial data to compute version"""

    def version_scheme(version):
        return VERSION.replace("D", ".dev0")

    def local_scheme(version):
        return "+{rev}.{distance}".format(rev=version.node[1:], distance=version.distance)

    return {"version_scheme": version_scheme, "local_scheme": local_scheme}


setup(
    name=NAME,
    version=VERSION,
    description="Web frontend for Libervia",
    long_description=long_description,
    author="Association « Salut à Toi »",
    author_email="contact@goffi.org",
    url="https://www.salut-a-toi.org",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Twisted",
        "License :: OSI Approved :: GNU Affero General Public License v3 "
        "or later (AGPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Topic :: Communications :: Chat",
    ],
    packages=["libervia", "libervia.common", "libervia.server", "twisted.plugins"],
    include_package_data=True,
    data_files=[(os.path.join("share", "doc", NAME), ["COPYING", "README", "INSTALL"])]
    + [
        (os.path.join("share", NAME, root), [os.path.join(root, f) for f in files])
        for root, dirs, files in os.walk("themes")
    ],
    entry_points={
        "console_scripts": [
            "libervia-web = libervia.server.launcher:Launcher.run",
            ],
        },
    zip_safe=False,
    setup_requires=["setuptools_scm"] if is_dev_version else [],
    use_scm_version=libervia_dev_version if is_dev_version else False,
    install_requires=install_requires,
    package_data={"libervia": ["VERSION"]},
    python_requires=">=3.7",
)



#!/usr/bin/env python3


from sat.core.i18n import _
from libervia.server.constants import Const as C
from twisted.internet import defer
from twisted.words.protocols.jabber import jid
from sat.tools.common import data_objects
from sat.core.log import getLogger

log = getLogger(__name__)

name = "chat_select"
access = C.PAGES_ACCESS_PROFILE
template = "chat/select.html"


@defer.inlineCallbacks
def prepare_render(self, request):
    profile = self.getProfile(request)
    template_data = request.template_data
    rooms = template_data["rooms"] = []
    bookmarks = yield self.host.bridgeCall("bookmarksList", "muc", "all", profile)
    for bm_values in list(bookmarks.values()):
        for room_jid, room_data in bm_values.items():
            url = self.getPageByName("chat").getURL(room_jid)
            rooms.append(data_objects.Room(room_jid, name=room_data.get("name"), url=url))
    rooms.sort(key=lambda r: r.name)


@defer.inlineCallbacks
def on_data_post(self, request):
    jid_ = self.getPostedData(request, "jid")
    if "@" not in jid_:
        profile = self.getProfile(request)
        service = yield self.host.bridgeCall("mucGetService", "", profile)
        if service:
            muc_jid = jid.JID(service)
            muc_jid.user = jid_
            jid_ = muc_jid.full()
        else:
            log.warning(_("Invalid jid received: {jid}".format(jid=jid_)))
            defer.returnValue(C.POST_NO_CONFIRM)
    url = self.getPageByName("chat").getURL(jid_)
    self.HTTPRedirect(request, url)



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from sat.core.i18n import _
from sat.tools.common import template_xmlui
from sat.tools.common import data_format
from sat.core.log import getLogger

"""merge-requests edition"""

name = "merge-requests_edit"
access = C.PAGES_ACCESS_PROFILE
template = "merge-request/edit.html"
log = getLogger(__name__)


def parse_url(self, request):
    try:
        item_id = self.nextPath(request)
    except IndexError:
        log.warning(_("no list item id specified"))
        self.pageError(request, C.HTTP_BAD_REQUEST)

    data = self.getRData(request)
    data["list_item_id"] = item_id


async def prepare_render(self, request):
    data = self.getRData(request)
    template_data = request.template_data
    service, node, list_item_id = (
        data.get("service", ""),
        data.get("node", ""),
        data["list_item_id"],
    )
    profile = self.getProfile(request)

    ignore = (
        "publisher",
        "author",
        "author_jid",
        "author_email",
        "created",
        "updated",
        "comments_uri",
        "request_data",
        "type",
    )
    merge_requests = data_format.deserialise(
        await self.host.bridgeCall(
            "mergeRequestsGet",
            service.full() if service else "",
            node,
            C.NO_LIMIT,
            [list_item_id],
            "",
            {},
            profile,
        )
    )
    list_item = template_xmlui.create(
        self.host, merge_requests['items'][0], ignore=ignore
    )

    try:
        # small trick to get a one line text input instead of the big textarea
        list_item.widgets["labels"].type = "string"
        list_item.widgets["labels"].value = list_item.widgets["labels"].value.replace(
            "\n", ", "
        )
    except KeyError:
        pass

    # same as list_edit
    wid = list_item.widgets['body']
    if wid.type == "xhtmlbox":
        wid.type = "textbox"
        wid.value =  await self.host.bridgeCall(
            "syntaxConvert", wid.value, C.SYNTAX_XHTML, "markdown",
            False, profile)

    template_data["new_list_item_xmlui"] = list_item


async def on_data_post(self, request):
    data = self.getRData(request)
    service = data["service"]
    node = data["node"]
    list_item_id = data["list_item_id"]
    posted_data = self.getAllPostedData(request)
    if not posted_data["title"] or not posted_data["body"]:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    try:
        posted_data["labels"] = [l.strip() for l in posted_data["labels"][0].split(",")]
    except (KeyError, IndexError):
        pass
    profile = self.getProfile(request)

    # we convert back body to XHTML
    body = await self.host.bridgeCall(
        "syntaxConvert", posted_data['body'][0], "markdown", C.SYNTAX_XHTML,
        False, profile)
    posted_data['body'] = ['<div xmlns="{ns}">{body}</div>'.format(ns=C.NS_XHTML,
                                                                     body=body)]

    extra = {'update': True}
    await self.host.bridgeCall(
        "mergeRequestSet",
        service.full(),
        node,
        "",
        "auto",
        posted_data,
        "",
        list_item_id,
        data_format.serialise(extra),
        profile,
    )
    # we don't want to redirect to edit page on success, but to list overview
    data["post_redirect_page"] = (
        self.getPageByName("merge-requests"),
        service.full(),
        node or "@",
    )



#!/usr/bin/env python3


# Libervia: a Salut à Toi frontend
# Copyright (C) 2011-2021 Jérôme Poisson <goffi@goffi.org>
# Copyright (C) 2013-2016 Adrien Cossa <souliane@mailoo.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sat.core.i18n import _, D_
from sat_frontends.tools.strings import addURLToText, fixXHTMLLinks
from sat.core.log import getLogger

log = getLogger(__name__)
from sat.tools.common import data_format
from sat.tools import xml_tools
from dbus.exceptions import DBusException
from twisted.internet import defer
from twisted.web import server
from twisted.web.resource import Resource
from twisted.words.protocols.jabber.jid import JID
from twisted.words.xish import domish
from jinja2 import Environment, PackageLoader
from datetime import datetime
import re
import os
import sys
import urllib.request, urllib.parse, urllib.error

from libervia.server.html_tools import sanitizeHtml, convertNewLinesToXHTML
from libervia.server.constants import Const as C

NS_ATOM = "http://www.w3.org/2005/Atom"
PARAMS_TO_GET = (
    C.STATIC_BLOG_PARAM_TITLE,
    C.STATIC_BLOG_PARAM_BANNER,
    C.STATIC_BLOG_PARAM_KEYWORDS,
    C.STATIC_BLOG_PARAM_DESCRIPTION,
)
re_strip_empty_div = re.compile(r"<div ?/>|<div> *?</div>")

# TODO: check disco features and use max_items when RSM is not available
# FIXME: change navigation links handling, this is is fragile
# XXX: this page will disappear, LiberviaPage will be used instead
# TODO: delete this page and create a compatibility page for links


def getDefaultQueryData(request):
    """Return query data which must be present in all links

    @param request(twisted.web.http.Request): request instance comming from render
    @return (dict): a dict with values as expected by urllib.urlencode
    """
    default_query_data = {}
    try:
        default_query_data["tag"] = request.extra_dict[
            "mam_filter_{}".format(C.MAM_FILTER_CATEGORY)
        ].encode("utf-8")
    except KeyError:
        pass
    return default_query_data


def _quote(value):
    """Quote a value for use in url

    @param value(unicode): value to quote
    @return (str): quoted value
    """
    return urllib.parse.quote(value.encode("utf-8"), "")


def _unquote(quoted_value):
    """Unquote a value coming from url

    @param unquote_value(str): value to unquote
    @return (unicode): unquoted value
    """
    assert not isinstance(quoted_value, str)
    return urllib.parse.unquote(quoted_value).decode("utf-8")


def _urlencode(query):
    """Same as urllib.urlencode, but use '&amp;' instead of '&'"""
    return "&amp;".join(
        [
            "{}={}".format(urllib.parse.quote_plus(str(k)), urllib.parse.quote_plus(str(v)))
            for k, v in query.items()
        ]
    )


class TemplateProcessor(object):

    THEME = "default"

    def __init__(self, host):
        self.host = host

        # add Libervia's themes directory to the python path
        sys.path.append(os.path.dirname(os.path.normpath(self.host.themes_dir)))
        themes = os.path.basename(os.path.normpath(self.host.themes_dir))
        self.env = Environment(loader=PackageLoader(themes, self.THEME))

    def useTemplate(self, request, tpl, data=None):
        theme_url = os.path.join("/", C.THEMES_URL, self.THEME)

        data_ = {
            "images": os.path.join(theme_url, "images"),
            "styles": os.path.join(theme_url, "styles"),
        }
        if data:
            data_.update(data)

        template = self.env.get_template("{}.html".format(tpl))
        return template.render(**data_).encode("utf-8")


class MicroBlog(Resource, TemplateProcessor):
    isLeaf = True

    def __init__(self, host):
        self.host = host
        Resource.__init__(self)
        TemplateProcessor.__init__(self, host)
        self.avatars_cache = {}

    def _avatarPathToUrl(self, avatar, request, bare_jid_s):
        filename = os.path.basename(avatar)
        avatar_url = os.path.join(self.host.service_cache_url, filename)
        self.avatars_cache[bare_jid_s] = avatar_url
        return avatar_url

    def getAvatarURL(self, pub_jid, request):
        """Return avatar of a jid if in cache, else ask for it.

        @param pub_jid (JID): publisher JID
        @return: deferred avatar URL (unicode)
        """
        bare_jid_s = pub_jid.userhost()
        try:
            url = self.avatars_cache[bare_jid_s]
        except KeyError:
            self.avatars_cache[
                bare_jid_s
            ] = ""  # avoid to request the vcard several times
            d = self.host.bridgeCall(
                "avatarGet", bare_jid_s, False, False, C.SERVICE_PROFILE
            )
            d.addCallback(self._avatarPathToUrl, request, bare_jid_s)
            return d
        return defer.succeed(url if url else C.DEFAULT_AVATAR_URL)

    def render_GET(self, request):
        if not request.postpath or len(request.postpath) > 2:
            return self.useTemplate(
                request, "static_blog_error", {"message": "You must indicate a nickname"}
            )

        prof_requested = _unquote(request.postpath[0])

        try:
            prof_found = self.host.bridge.profileNameGet(prof_requested)
        except DBusException:
            prof_found = None
        if not prof_found or prof_found == C.SERVICE_PROFILE:
            return self.useTemplate(
                request, "static_blog_error", {"message": "Invalid nickname"}
            )

        d = defer.Deferred()
        # TODO: jid caching
        self.host.bridge.asyncGetParamA(
            "JabberID",
            "Connection",
            "value",
            profile_key=prof_found,
            callback=d.callback,
            errback=d.errback,
        )
        d.addCallback(self.render_gotJID, request, prof_found)
        return server.NOT_DONE_YET

    def render_gotJID(self, pub_jid_s, request, profile):
        pub_jid = JID(pub_jid_s)

        request.extra_dict = {}  # will be used for RSM and MAM
        self.parseURLParams(request)
        if request.item_id:
            # FIXME: this part seems useless
            # we want a specific item
            # item_ids = [request.item_id]
            # max_items = 1
            max_items = C.NO_LIMIT  # FIXME
        else:
            # max_items = int(request.extra_dict['rsm_max']) # FIXME
            max_items = C.NO_LIMIT
            # TODO: use max_items only when RSM is not available

        if request.atom:
            request.extra_dict.update(request.mam_extra)
            self.getAtom(
                pub_jid,
                max_items,
                request.extra_dict,
                request.extra_comments_dict,
                request,
                profile,
            )

        elif request.item_id:
            # we can't merge mam_extra now because we'll use item_ids
            self.getItemById(
                pub_jid,
                request.item_id,
                request.extra_dict,
                request.extra_comments_dict,
                request,
                profile,
            )
        else:
            request.extra_dict.update(request.mam_extra)
            self.getItems(
                pub_jid,
                max_items,
                request.extra_dict,
                request.extra_comments_dict,
                request,
                profile,
            )

    ## URL parsing

    def parseURLParams(self, request):
        """Parse the request URL parameters.

        @param request: HTTP request
        """
        if len(request.postpath) > 1:
            if request.postpath[1] == "atom.xml":  # return the atom feed
                request.atom = True
                request.item_id = None
            else:
                request.atom = False
                request.item_id = _unquote(request.postpath[1])
        else:
            request.item_id = None
            request.atom = False

        self.parseURLParamsRSM(request)
        # XXX: request.display_single is True when only one blog post is visible
        request.display_single = (request.item_id is not None) or int(
            request.extra_dict["rsm_max"]
        ) == 1
        self.parseURLParamsCommentsRSM(request)
        self.parseURLParamsMAM(request)

    def parseURLParamsRSM(self, request):
        """Parse RSM request data from the URL parameters for main items

        fill request.extra_dict accordingly
        @param request: HTTP request
        """
        if request.item_id:  # XXX: item_id and RSM are not compatible
            return
        try:
            rsm_max = int(request.args["max"][0])
            if rsm_max > C.STATIC_RSM_MAX_LIMIT:
                log.warning("Request with rsm_max over limit ({})".format(rsm_max))
                rsm_max = C.STATIC_RSM_MAX_LIMIT
            request.extra_dict["rsm_max"] = str(rsm_max)
        except (ValueError, KeyError):
            request.extra_dict["rsm_max"] = str(C.STATIC_RSM_MAX_DEFAULT)
        try:
            request.extra_dict["rsm_index"] = request.args["index"][0]
        except (ValueError, KeyError):
            try:
                request.extra_dict["rsm_before"] = request.args["before"][0].decode(
                    "utf-8"
                )
            except KeyError:
                try:
                    request.extra_dict["rsm_after"] = request.args["after"][0].decode(
                        "utf-8"
                    )
                except KeyError:
                    pass

    def parseURLParamsCommentsRSM(self, request):
        """Parse RSM request data from the URL parameters for comments

        fill request.extra_dict accordingly
        @param request: HTTP request
        """
        request.extra_comments_dict = {}
        if request.display_single:
            try:
                rsm_max = int(request.args["comments_max"][0])
                if rsm_max > C.STATIC_RSM_MAX_LIMIT:
                    log.warning("Request with rsm_max over limit ({})".format(rsm_max))
                    rsm_max = C.STATIC_RSM_MAX_LIMIT
                request.extra_comments_dict["rsm_max"] = str(rsm_max)
            except (ValueError, KeyError):
                request.extra_comments_dict["rsm_max"] = str(
                    C.STATIC_RSM_MAX_COMMENTS_DEFAULT
                )
        else:
            request.extra_comments_dict["rsm_max"] = "0"

    def parseURLParamsMAM(self, request):
        """Parse MAM request data from the URL parameters for main items

        fill request.extra_dict accordingly
        @param request: HTTP request
        """
        # XXX: we use a separate dict for MAM as the filters are not used
        #      when display_single is set (because it then use item_ids which
        #      can't be used with MAM), but it is still used in this case
        #      for navigation links.
        request.mam_extra = {}
        try:
            request.mam_extra[
                "mam_filter_{}".format(C.MAM_FILTER_CATEGORY)
            ] = request.args["tag"][0].decode("utf-8")
        except KeyError:
            pass

    ## Items retrieval

    def getItemById(
        self, pub_jid, item_id, extra_dict, extra_comments_dict, request, profile
    ):
        """

        @param pub_jid (jid.JID): publisher JID
        @param item_id(unicode): ID of the item to retrieve
        @param extra_dict (dict): extra configuration for initial items only
        @param extra_comments_dict (dict): extra configuration for comments only
        @param request: HTTP request
        @param profile
        """

        def gotItems(items):
            items, metadata = items
            items = [data_format.deserialise(i) for i in items]
            item = items[0]  # assume there's only one item

            def gotMetadata(result):
                __, rsm_metadata = result
                try:
                    metadata["rsm_count"] = rsm_metadata["rsm_count"]
                except KeyError:
                    pass
                try:
                    metadata["rsm_index"] = str(int(rsm_metadata["rsm_index"]) - 1)
                except KeyError:
                    pass

                metadata["rsm_first"] = metadata["rsm_last"] = item["id"]

                def gotComments(comments):
                    # at this point we can merge mam dict
                    request.extra_dict.update(request.mam_extra)
                    # build the items as self.getItems would do it (and as self.renderHTML expects them to be)
                    comments = [
                        (
                            item["comments_service"],
                            item["comments_node"],
                            "",
                            [data_format.deserialise(c) for c in comments[0]],
                            comments[1],
                        )
                    ]
                    self.renderHTML(
                        [(item, comments)], metadata, request, pub_jid, profile
                    )

                # get the comments
                # max_comments = int(extra_comments_dict['rsm_max']) # FIXME
                max_comments = C.NO_LIMIT
                # TODO: use max_comments only when RSM is not available
                self.host.bridge.mbGet(
                    item["comments_service"],
                    item["comments_node"],
                    max_comments,
                    [],
                    extra_comments_dict,
                    C.SERVICE_PROFILE,
                    callback=gotComments,
                    errback=lambda failure: self.renderError(failure, request, pub_jid),
                )

            # XXX: retrieve RSM information related to the main item. We can't do it while
            # retrieving the item, because item_ids and rsm should not be used together.
            self.host.bridge.mbGet(
                pub_jid.userhost(),
                "",
                0,
                [],
                {"rsm_max": "1", "rsm_after": item["id"]},
                C.SERVICE_PROFILE,
                callback=gotMetadata,
                errback=lambda failure: self.renderError(failure, request, pub_jid),
            )

        # get the main item
        self.host.bridge.mbGet(
            pub_jid.userhost(),
            "",
            0,
            [item_id],
            extra_dict,
            C.SERVICE_PROFILE,
            callback=gotItems,
            errback=lambda failure: self.renderError(failure, request, pub_jid),
        )

    def getItems(
        self, pub_jid, max_items, extra_dict, extra_comments_dict, request, profile
    ):
        """

        @param pub_jid (jid.JID): publisher JID
        @param max_items(int): maximum number of item to get, C.NO_LIMIT for no limit
        @param extra_dict (dict): extra configuration for initial items only
        @param extra_comments_dict (dict): extra configuration for comments only
        @param request: HTTP request
        @param profile
        """

        def getResultCb(data, rt_session):
            remaining, results = data
            # we have requested one node only
            assert remaining == 0
            assert len(results) == 1
            service, node, failure, items, metadata = results[0]
            items = [(data_format.deserialise(i), m) for i,m in items]
            if failure:
                self.renderError(failure, request, pub_jid)
            else:
                self.renderHTML(items, metadata, request, pub_jid, profile)

        def getResult(rt_session):
            self.host.bridge.mbGetFromManyWithCommentsRTResult(
                rt_session,
                C.SERVICE_PROFILE,
                callback=lambda data: getResultCb(data, rt_session),
                errback=lambda failure: self.renderError(failure, request, pub_jid),
            )

        # max_comments = int(extra_comments_dict['rsm_max']) # FIXME
        max_comments = 0
        # TODO: use max_comments only when RSM is not available
        self.host.bridge.mbGetFromManyWithComments(
            C.JID,
            [pub_jid.userhost()],
            max_items,
            max_comments,
            extra_dict,
            extra_comments_dict,
            C.SERVICE_PROFILE,
            callback=getResult,
        )

    def getAtom(
        self, pub_jid, max_items, extra_dict, extra_comments_dict, request, profile
    ):
        """

        @param pub_jid (jid.JID): publisher JID
        @param max_items(int): maximum number of item to get, C.NO_LIMIT for no limit
        @param extra_dict (dict): extra configuration for initial items only
        @param extra_comments_dict (dict): extra configuration for comments only
        @param request: HTTP request
        @param profile
        """

        def gotItems(data):
            # Generate a clean atom feed with uri linking to this blog
            # from microblog data
            items, metadata = data
            items = [data_format.deserialise(i) for i in items]
            feed_elt = domish.Element((NS_ATOM, "feed"))
            title = _("{user}'s blog").format(user=profile)
            feed_elt.addElement("title", content=title)

            base_blog_url = self.host.getExtBaseURL(
                request, "blog/{user}".format(user=profile)
            )

            # atom link
            link_feed_elt = feed_elt.addElement("link")
            link_feed_elt["href"] = "{base}/atom.xml".format(base=base_blog_url)
            link_feed_elt["type"] = "application/atom+xml"
            link_feed_elt["rel"] = "self"

            # blog link
            link_blog_elt = feed_elt.addElement("link")
            link_blog_elt["rel"] = "alternate"
            link_blog_elt["type"] = "text/html"
            link_blog_elt["href"] = base_blog_url

            # blog link XMPP uri
            blog_xmpp_uri = metadata["uri"]
            link_blog_elt = feed_elt.addElement("link")
            link_blog_elt["rel"] = "alternate"
            link_blog_elt["type"] = "application/atom+xml"
            link_blog_elt["href"] = blog_xmpp_uri

            feed_elt.addElement("id", content=_quote(blog_xmpp_uri))
            updated_unix = max([float(item["updated"]) for item in items])
            updated_dt = datetime.fromtimestamp(updated_unix)
            feed_elt.addElement(
                "updated", content="{}Z".format(updated_dt.isoformat("T"))
            )

            for item in items:
                entry_elt = feed_elt.addElement("entry")

                # Title
                try:
                    title = item["title"]
                except KeyError:
                    # for microblog (without title), we use an abstract of content as title
                    title = "{}…".format(" ".join(item["content"][:70].split()))
                entry_elt.addElement("title", content=title)

                # HTTP link
                http_link_elt = entry_elt.addElement("link")
                http_link_elt["rel"] = "alternate"
                http_link_elt["type"] = "text/html"
                http_link_elt["href"] = "{base}/{quoted_id}".format(
                    base=base_blog_url, quoted_id=_quote(item["id"])
                )
                # XMPP link
                xmpp_link_elt = entry_elt.addElement("link")
                xmpp_link_elt["rel"] = "alternate"
                xmpp_link_elt["type"] = "application/atom+xml"
                xmpp_link_elt["href"] = "{blog_uri};item={item_id}".format(
                    blog_uri=blog_xmpp_uri, item_id=item["id"]
                )

                # date metadata
                entry_elt.addElement("id", content=item["atom_id"])
                updated = datetime.fromtimestamp(float(item["updated"]))
                entry_elt.addElement(
                    "updated", content="{}Z".format(updated.isoformat("T"))
                )
                published = datetime.fromtimestamp(float(item["published"]))
                entry_elt.addElement(
                    "published", content="{}Z".format(published.isoformat("T"))
                )

                # author metadata
                author_elt = entry_elt.addElement("author")
                author_elt.addElement("name", content=item.get("author", profile))
                try:
                    author_elt.addElement(
                        "uri", content="xmpp:{}".format(item["author_jid"])
                    )
                except KeyError:
                    pass
                try:
                    author_elt.addElement("email", content=item["author_email"])
                except KeyError:
                    pass

                # categories
                for tag in item.get('tags', []):
                    category_elt = entry_elt.addElement("category")
                    category_elt["term"] = tag

                # content
                try:
                    content_xhtml = item["content_xhtml"]
                except KeyError:
                    content_elt = entry_elt.addElement("content", content="content")
                    content_elt["type"] = "text"
                else:
                    content_elt = entry_elt.addElement("content")
                    content_elt["type"] = "xhtml"
                    content_elt.addChild(
                        xml_tools.ElementParser()(content_xhtml, namespace=C.NS_XHTML)
                    )

            atom_feed = '<?xml version="1.0" encoding="utf-8"?>\n{}'.format(
                feed_elt.toXml()
            )
            self.renderAtomFeed(atom_feed, request),

        self.host.bridge.mbGet(
            pub_jid.userhost(),
            "",
            max_items,
            [],
            extra_dict,
            C.SERVICE_PROFILE,
            callback=gotItems,
        )

    ## rendering

    def _updateDict(self, value, dict_, key):
        dict_[key] = value

    def _getImageParams(self, options, key, default, alt):
        """regexp from http://answers.oreilly.com/topic/280-how-to-validate-urls-with-regular-expressions/"""
        url = options[key] if key in options else ""
        regexp = (
            r"^(https?|ftp)://[a-z0-9-]+(\.[a-z0-9-]+)+(/[\w-]+)*/[\w-]+\.(gif|png|jpg)$"
        )
        if re.match(regexp, url):
            url = url
        else:
            url = default
        return BlogImage(url, alt)

    def renderError(self, failure, request, pub_jid):
        request.setResponseCode(500)
        request.write(
            self.useTemplate(
                request, "static_blog_error", {"message": "Can't access requested data"}
            )
        )
        request.finish()

    def renderHTML(self, items, metadata, request, pub_jid, profile):
        """Retrieve the user parameters before actually rendering the static blog

        @param items(list[tuple(dict, list)]): same as in self._renderHTML
        @param metadata(dict): original node metadata
        @param request: HTTP request
        @param pub_jid (JID): publisher JID
        @param profile (unicode): %(doc_profile)s
        """
        d_list = []
        options = {}

        d = self.getAvatarURL(pub_jid, request)
        d.addCallback(self._updateDict, options, "avatar")
        d.addErrback(self.renderError, request, pub_jid)
        d_list.append(d)

        for param_name in PARAMS_TO_GET:
            d = defer.Deferred()
            self.host.bridge.asyncGetParamA(
                param_name,
                C.STATIC_BLOG_KEY,
                "value",
                C.SERVER_SECURITY_LIMIT,
                profile,
                callback=d.callback,
                errback=d.errback,
            )
            d.addCallback(self._updateDict, options, param_name)
            d.addErrback(self.renderError, request, pub_jid)
            d_list.append(d)

        dlist_d = defer.DeferredList(d_list)
        dlist_d.addCallback(
            lambda __: self._renderHTML(items, metadata, options, request, pub_jid)
        )

    def _renderHTML(self, items, metadata, options, request, pub_jid):
        """Actually render the static blog.

        If mblog_data is a list of dict, we are missing the comments items so we just
        display the main items. If mblog_data is a list of couple, each couple is
        associating a main item data with the list of its comments, so we render all.
        @param items(list[tuple(dict, list)]): list of 2-tuple with
            - item(dict): item microblog data
            - comments_list(list[tuple]): list of 5-tuple with
                - service (unicode): pubsub service where the comments node is
                - node (unicode): comments node
                - failure (unicode): empty in case of success, else error message
                - comments(list[dict]): list of microblog data
                - comments_metadata(dict): metadata of the comment node
        @param metadata(dict): original node metadata
        @param options: dict defining the blog's parameters
        @param request: the HTTP request
        @param pub_jid (JID): publisher JID
        """
        if not isinstance(options, dict):
            options = {}
        user = sanitizeHtml(pub_jid.user)
        base_url = os.path.join("/blog/", user)

        def getOption(key):
            return sanitizeHtml(options[key]) if key in options else ""

        avatar = os.path.normpath("/{}".format(getOption("avatar")))
        title = getOption(C.STATIC_BLOG_PARAM_TITLE) or user
        query_data = _urlencode(getDefaultQueryData(request)).decode("utf-8")

        xmpp_uri = metadata["uri"]
        if len(items) == 1:
            # FIXME: that's really not a good way to get item id
            #        this must be changed after static blog refactorisation
            item_id = items[0][0]["id"]
            xmpp_uri += ";item={}".format(_quote(item_id))

        data = {
            "url_base": base_url,
            "xmpp_uri": xmpp_uri,
            "url_query": "?{}".format(query_data) if query_data else "",
            "keywords": getOption(C.STATIC_BLOG_PARAM_KEYWORDS),
            "description": getOption(C.STATIC_BLOG_PARAM_DESCRIPTION),
            "title": title,
            "favicon": avatar,
            "banner_img": self._getImageParams(
                options, C.STATIC_BLOG_PARAM_BANNER, avatar, title
            ),
        }

        data["navlinks"] = NavigationLinks(request, items, metadata, base_url)
        data["messages"] = []
        for item in items:
            item, comments_list = item
            comments, comments_count = [], 0
            for node_comments in comments_list:
                comments.extend(node_comments[3])
                try:
                    comments_count += int(node_comments[4]["rsm_count"])
                except KeyError:
                    pass
            data["messages"].append(
                BlogMessage(request, base_url, item, comments, comments_count)
            )

        request.write(self.useTemplate(request, "static_blog", data))
        request.finish()

    def renderAtomFeed(self, feed, request):
        request.write(feed.encode("utf-8"))
        request.finish()


class NavigationLinks(object):
    def __init__(self, request, items, metadata, base_url):
        """Build the navigation links.

        @param items (list): list of items
        @param metadata (dict): rsm data
        @param base_url (unicode): the base URL for this user's blog
        @return: dict
        """
        # FIXME: this code must be refactorized, it is fragile
        #        and difficult to maintain

        # query data which must be present in all links
        default_query_data = getDefaultQueryData(request)

        # which links we need to display
        if request.display_single:
            links = ("later_message", "older_message")
            # key must exist when using the template
            self.later_messages = self.older_messages = ""
        else:
            links = ("later_messages", "older_messages")
            self.later_message = self.older_message = ""

        # now we set the links according to RSM
        for key in links:
            query_data = default_query_data.copy()

            if key.startswith("later_message"):
                try:
                    index = int(metadata["rsm_index"])
                except (KeyError, ValueError):
                    pass
                else:
                    if index == 0:
                        # we don't show this link on first page
                        setattr(self, key, "")
                        continue
                try:
                    query_data["before"] = metadata["rsm_first"].encode("utf-8")
                except KeyError:
                    pass
            else:
                try:
                    index = int(metadata["rsm_index"])
                    count = int(metadata.get("rsm_count"))
                except (KeyError, ValueError):
                    # XXX: if we don't have index or count, we can't know if we
                    #      are on the last page or not
                    pass
                else:
                    # if we have index, we don't show the after link
                    # on the last page
                    if index + len(items) >= count:
                        setattr(self, key, "")
                        continue
                try:
                    query_data["after"] = metadata["rsm_last"].encode("utf-8")
                except KeyError:
                    pass

            if request.display_single:
                query_data["max"] = 1

            link = "{}?{}".format(base_url, _urlencode(query_data))
            setattr(self, key, BlogLink(link, key, key.replace("_", " ")))


class BlogImage(object):
    def __init__(self, url_, alt):
        self.url = url_
        self.alt = alt


class BlogLink(object):
    def __init__(self, url_, style, text):
        self.url = url_
        self.style = style
        self.text = text


class BlogMessage(object):
    def __init__(self, request, base_url, entry, comments=None, comments_count=0):
        """

        @param request: HTTP request
        @param base_url (unicode): the base URL
        @param entry(dict): item microblog data
        @param comments(list[dict]): list of microblog data
        @param comments_count (int): total number of comments
        """
        if comments is None:
            comments = []
        timestamp = float(entry.get("published", 0))

        # FIXME: for now we assume that the comments' depth is only 1
        is_comment = not entry.get("comments", False)

        self.date = datetime.fromtimestamp(timestamp)
        self.type = "comment" if is_comment else "main_item"
        self.style = "mblog_comment" if is_comment else ""
        self.content = self.getText(entry, "content")

        if is_comment:
            self.author = _("from {}").format(entry["author"])
        else:
            self.author = "&nbsp;"
            self.url = "{}/{}".format(base_url, _quote(entry["id"]))
            query_data = getDefaultQueryData(request)
            if query_data:
                self.url += "?{}".format(_urlencode(query_data))
            self.title = self.getText(entry, "title")
            self.tags = [sanitizeHtml(tag) for tag in entry.get('tags', [])]

            count_text = lambda count: D_("comments") if count > 1 else D_("comment")

            self.comments_text = "{} {}".format(
                comments_count, count_text(comments_count)
            )

            delta = comments_count - len(comments)
            if request.display_single and delta > 0:
                prev_url = "{}?{}".format(
                    self.url, _urlencode({"comments_max": comments_count})
                )
                prev_text = D_("show {count} previous {comments}").format(
                    count=delta, comments=count_text(delta)
                )
                self.all_comments_link = BlogLink(prev_url, "comments_link", prev_text)

        if comments:
            self.comments = [
                BlogMessage(request, base_url, comment) for comment in comments
            ]

    def getText(self, entry, key):
        try:
            xhtml = entry["{}_xhtml".format(key)]
        except KeyError:
            try:
                processor = addURLToText if key.startswith("content") else sanitizeHtml
                return convertNewLinesToXHTML(processor(entry[key]))
            except KeyError:
                return None
        else:
            # FIXME: empty <div /> elements provoke rendering issue
            #        this regex is a temporary workadound, need more investigation
            xhtml = re_strip_empty_div.sub("", xhtml)
            return fixXHTMLLinks(xhtml)



#!/usr/bin/env python3


name = "user_blog"


def parse_url(self, request):
    # in this subpage, we want path args and query args
    # (i.e. what's remaining in URL: filters, id, etc.)
    # to be used by blog's url parser, so we don't skip parse_url
    data = self.getRData(request)
    service = data["service"]
    self.pageRedirect(
        "blog_view", request, skip_parse_url=False, path_args=[service.full(), "@"]
    )



#!/usr/bin/env python3

from libervia.server.constants import Const as C
from twisted.internet import defer
from sat.core.log import getLogger
from sat.core.i18n import D_
from sat.core import exceptions
from sat_frontends.bridge.bridge_frontend import BridgeException

"""creation of new events"""

name = "photos_new"
access = C.PAGES_ACCESS_PROFILE
template = "photo/create.html"
log = getLogger(__name__)


async def on_data_post(self, request):
    request_data = self.getRData(request)
    profile = self.getProfile(request)
    name = self.getPostedData(request, "name").replace('/', '_')
    albums_path = "/albums"
    album_path = f"{albums_path}/{name}"
    if profile is None:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    fis_ns = self.host.ns_map["fis"]
    http_upload_ns = self.host.ns_map["http_upload"]
    entities_services, __, __ = await self.host.bridgeCall(
        "discoFindByFeatures",
        [fis_ns, http_upload_ns],
        [],
        False,
        True,
        False,
        False,
        False,
        profile
    )
    try:
        fis_service = next(iter(entities_services))
    except StopIteration:
        raise exceptions.DataError(D_(
           "You server has no service to create a photo album, please ask your server "
           "administrator to add one"))

    try:
        await self.host.bridgeCall(
            "FISCreateDir",
            fis_service,
            "",
            albums_path,
            {"access_model": "open"},
            profile
        )
    except BridgeException as e:
        if e.condition == 'conflict':
            pass
        else:
            log.error(f"Can't create {albums_path} path: {e}")
            raise e

    try:
        await self.host.bridgeCall(
            "FISCreateDir",
            fis_service,
            "",
            album_path,
            {"access_model": "whitelist"},
            profile
        )
    except BridgeException as e:
        if e.condition == 'conflict':
            pass
        else:
            log.error(f"Can't create {album_path} path: {e}")
            raise e

    await self.host.bridgeCall(
        "interestsRegisterFileSharing",
        fis_service,
        "photos",
        "",
        album_path,
        name,
        "",
        profile
    )
    log.info(f"album {name} created")
    request_data["post_redirect_page"] = self.getPageByName("photos")
    defer.returnValue(C.POST_NO_CONFIRM)



#!/ur/bin/env python3

import shutil
import json
from pathlib import Path
from ast import literal_eval
from sat.core.i18n import _
from sat.core.log import getLogger
from sat.core import exceptions
from sat.tools.common import utils
from libervia.server.constants import Const as C
from libervia.server.classes import Script
from libervia.server.tasks import task


log = getLogger(__name__)


class Task(task.Task):

    def prepare(self):
        if "brython" not in self.resource.browser_modules:
            raise exceptions.CancelError("No brython module found")

        brython_js = self.build_path / "brython.js"
        if not brython_js.is_file():
            installed_ver = None
        else:
            with brython_js.open() as f:
                for line in f:
                    if line.startswith('// implementation ['):
                        installed_ver = literal_eval(line[18:])[:3]
                        log.debug(
                            f"brython v{'.'.join(str(v) for v in installed_ver)} already "
                            f"installed")
                        break
                else:
                    log.warning(
                        f"brython file at {brython_js} doesn't has implementation "
                        f"version"
                    )
                    installed_ver = None

        try:
            import brython
        except ModuleNotFoundError as e:
            log.error('"brython" module is missing, can\'t use browser code for Brython')
            raise e
        ver = [int(v) for v in brython.implementation.split('.')[:3]]
        if ver != installed_ver:
            log.info(_("Installing Brython v{version}").format(
                version='.'.join(str(v) for v in ver)))
            data_path = Path(brython.__file__).parent / 'data'
            # shutil has blocking method, but the task is run before we start
            # the web server, so it's not a big deal
            shutil.copyfile(data_path / "brython.js", brython_js)
            shutil.copy(data_path / "brython_stdlib.js", self.build_path)
        else:
            log.debug("Brython is already installed")

        self.WATCH_DIRS = []
        self.setCommonScripts()

    def setCommonScripts(self):
        for dyn_data in self.resource.browser_modules["brython"]:
            url_hash = dyn_data['url_hash']
            import_url = f"/{C.BUILD_DIR}/{C.BUILD_DIR_DYN}/{url_hash}"
            dyn_data.setdefault('scripts', utils.OrderedSet()).update([
                Script(src=f"/{C.BUILD_DIR}/brython.js"),
                Script(src=f"/{C.BUILD_DIR}/brython_stdlib.js"),
            ])
            dyn_data.setdefault('template', {})['body_onload'] = self.getBodyOnload(
                extra_path=[import_url])
            self.WATCH_DIRS.append(dyn_data['path'].resolve())

    def getBodyOnload(self, debug=True, cache=True, extra_path=None):
            on_load_opts = {"pythonpath": [f"/{C.BUILD_DIR}"]}
            if debug:
                on_load_opts[debug] = 1
            if cache:
                on_load_opts["cache"] = True
            if extra_path is not None:
                on_load_opts["pythonpath"].extend(extra_path)

            return f"brython({json.dumps(on_load_opts)})"

    def copyFiles(self, files_paths, dest):
        for p in files_paths:
            log.debug(f"copying {p}")
            if p.is_dir():
                if p.name == '__pycache__':
                    continue
                shutil.copytree(p, dest / p.name)
            else:
                shutil.copy(p, dest)

    async def onDirEvent(self, host, filepath, flags):
        self.setCommonScripts()
        await self.manager.runTaskInstance(self)

    def start(self):
        dyn_path = self.build_path / C.BUILD_DIR_DYN
        for dyn_data in self.resource.browser_modules["brython"]:
            url_hash = dyn_data['url_hash']
            if url_hash is None:
                # root modules
                url_prefix = dyn_data.get('url_prefix')
                if url_prefix is None:
                    dest = self.build_path
                    init_dest_url = f"/{C.BUILD_DIR}/__init__.py"
                else:
                    dest = self.build_path / url_prefix
                    dest.mkdir(exist_ok = True)
                    init_dest_url = f"/{C.BUILD_DIR}/{url_prefix}/__init__.py"

                self.copyFiles(dyn_data['path'].glob('*py'), dest)

                init_file = dyn_data['path'] / '__init__.py'
                if init_file.is_file():
                    self.resource.dyn_data_common['scripts'].update([
                        Script(src=f"/{C.BUILD_DIR}/brython.js"),
                        Script(type='text/python', src=init_dest_url)
                    ])
                    self.resource.dyn_data_common.setdefault(
                        "template", {})['body_onload'] = self.getBodyOnload()
            else:
                page_dyn_path = dyn_path / url_hash
                log.debug(f"using dynamic path at {page_dyn_path}")
                if page_dyn_path.exists():
                    log.debug("cleaning existing path")
                    shutil.rmtree(page_dyn_path)

                page_dyn_path.mkdir(parents=True, exist_ok=True)
                log.debug("copying browser python files")
                self.copyFiles(dyn_data['path'].iterdir(), page_dyn_path)

                script = Script(
                    type='text/python',
                    src=f"/{C.BUILD_DIR}/{C.BUILD_DIR_DYN}/{url_hash}/__init__.py"
                )
                dyn_data.setdefault('scripts', utils.OrderedSet()).add(script)



"""text edition management"""

from browser import document, window, aio, bind, timer
from browser.local_storage import storage
from browser.object_storage import ObjectStorage
from javascript import JSON
from aio_bridge import Bridge, BridgeException
from template import Template
import dialog

bridge = Bridge()
object_storage = ObjectStorage(storage)
profile = window.profile

# how often we save forms, in seconds
AUTOSAVE_FREQUENCY = 20


def serialise_form(form_elt):
    ret = {}
    for elt in form_elt.elements:
        if elt.tagName == "INPUT":
            if elt.type in ("hidden", "submit"):
                continue
            elif elt.type == "text":
                ret[elt.name] = elt.value
            else:
                print(f"elt.type not managet yet: {elt.type}")
                continue
        elif elt.tagName == "TEXTAREA":
            ret[elt.name] = elt.value
        elif elt.tagName in ("BUTTON",):
            continue
        else:
            print(f"tag not managet yet: {elt.tagName}")
            continue
    return ret


def restore_form(form_elt, data):
    for elt in form_elt.elements:
        if elt.tagName not in ("INPUT", "TEXTAREA"):
            continue
        try:
            value = data[elt.name]
        except KeyError:
            continue
        else:
            elt.value = value


def set_form_autosave(form_id):
    """Save locally form data regularly and restore it until it's submitted

    form is saved every AUTOSAVE_FREQUENCY seconds and when visibility is lost.
    Saved data is restored when the method is called.
    Saved data is cleared when the form is submitted.
    """
    if profile is None:
        print(f"No session started, won't save and restore form {form_id}")
        return

    form_elt = document[form_id]
    submitted = False

    key = {"profile": profile, "type": "form_autosave", "form": form_id}
    try:
        form_saved_data = object_storage[key]
    except KeyError:
        last_serialised = None
    else:
        print(f"restoring content of form {form_id!r}")
        last_serialised = form_saved_data
        restore_form(form_elt, form_saved_data)

    def save_form():
        if not submitted:
            nonlocal last_serialised
            serialised = serialise_form(form_elt)
            if serialised != last_serialised:
                last_serialised = serialised
                print(f"saving content of form {form_id!r}")
                object_storage[key] = serialised

    @bind(form_elt, "submit")
    def on_submit(evt):
        nonlocal submitted
        submitted = True
        print(f"clearing stored content of form {form_id!r}")
        try:
            del object_storage[key]
        except KeyError:
            print("key error")
            pass

    @bind(document, "visibilitychange")
    def on_visibiliy_change(evt):
        print("visibility change")
        if document.visibilityState != "visible":
            save_form()

    timer.set_interval(save_form, AUTOSAVE_FREQUENCY * 1000)


class TagsEditor:

    def __init__(self, input_selector):
        print("installing Tags Editor")
        self.input_elt = document.select_one(input_selector)
        self.input_elt.style.display = "none"
        tags_editor_tpl = Template('editor/tags_editor.html')
        self.tag_tpl = Template('editor/tag.html')

        editor_elt = tags_editor_tpl.get_elt()
        self.input_elt.parent <= editor_elt
        self.tag_input_elt = editor_elt.select_one(".tag_input")
        self.tag_input_elt.bind("keydown", self.on_key_down)
        self._current_tags = None
        self.tags_map = {}
        for tag in self.current_tags:
            self.add_tag(tag, force=True)

    @property
    def current_tags(self):
        if self._current_tags is None:
            self._current_tags = {
                t.strip() for t in self.input_elt.value.split(',') if t.strip()
            }
        return self._current_tags

    @current_tags.setter
    def current_tags(self, tags):
        self._current_tags = tags

    def add_tag(self, tag, force=False):
        tag = tag.strip()
        if not force and (not tag or tag in self.current_tags):
            return
        self.current_tags = self.current_tags | {tag}
        self.input_elt.value = ','.join(self.current_tags)
        tag_elt = self.tag_tpl.get_elt({"label": tag})
        self.tags_map[tag] = tag_elt
        self.tag_input_elt.parent.insertBefore(tag_elt, self.tag_input_elt)
        tag_elt.select_one(".click_to_delete").bind(
            "click", lambda evt: self.on_tag_click(evt, tag)
        )

    def remove_tag(self, tag):
        try:
            tag_elt = self.tags_map[tag]
        except KeyError:
            print(f"trying to remove an inexistant tag: {tag}")
        else:
            self.current_tags = self.current_tags - {tag}
            self.input_elt.value = ','.join(self.current_tags)
            tag_elt.remove()

    def on_tag_click(self, evt, tag):
        evt.stopPropagation()
        self.remove_tag(tag)

    def on_key_down(self, evt):
        if evt.key in (",", "Enter"):
            evt.stopPropagation()
            evt.preventDefault()
            self.add_tag(self.tag_input_elt.value)
            self.tag_input_elt.value = ""


class BlogEditor:
    """Editor class, handling tabs, preview, and submit loading button

    It's using and HTML form as source
    The form must have:
        - a "title" text input
        - a "body" textarea
        - an optional "tags" text input with comma separated tags (may be using Tags
          Editor)
        - a "tab_preview" tab element
    """

    def __init__(self, form_id="blog_post_edit"):
        self.tab_select = window.tab_select
        self.item_tpl = Template('blog/item.html')
        self.form = document[form_id]
        for elt in document.select(".click_to_edit"):
            elt.bind("click", self.on_edit)
        for elt in document.select('.click_to_preview'):
            elt.bind("click", lambda evt: aio.run(self.on_preview(evt)))
        self.form.bind("submit", self.on_submit)


    def on_edit(self, evt):
        self.tab_select(evt.target, "tab_edit", "is-active")

    async def on_preview(self, evt):
        """Generate a blog preview from HTML form

        """
        print("on preview OK")
        elt = evt.target
        tab_preview = document["tab_preview"]
        tab_preview.clear()
        data = {
            "content_rich": self.form.select_one('textarea[name="body"]').value.strip()
        }
        title = self.form.select_one('input[name="title"]').value.strip()
        if title:
            data["title_rich"] = title
        tags_input_elt = self.form.select_one('input[name="tags"]')
        if tags_input_elt is not None:
            tags = tags_input_elt.value.strip()
            if tags:
                data['tags'] = [t.strip() for t in tags.split(',') if t.strip()]
        try:
            preview_data = JSON.parse(
                await bridge.mbPreview("", "", JSON.stringify(data))
            )
        except BridgeException as e:
            dialog.notification.show(
                f"Can't generate item preview: {e.message}",
                level="error"
            )
        else:
            self.tab_select(elt, "tab_preview", "is-active")
            item_elt = self.item_tpl.get_elt({
                "item": preview_data,
                "dates_format": "short",
            })
            tab_preview <= item_elt

    def on_submit(self, evt):
        submit_btn = document.select_one("button[type='submit']")
        submit_btn.classList.add("is-loading")



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from sat.core.i18n import D_
from sat.core.log import getLogger

log = getLogger(__name__)

name = "forum_topic_new"
label = D_("New Topic")
access = C.PAGES_ACCESS_PROFILE
template = "blog/publish.html"


async def prepare_render(self, request):
   template_data = request.template_data
   template_data.update({
       "post_form_id": "forum_topic_edit",
       "publish_title": D_("New Forum Topic"),
       "title_label": D_("Topic"),
       "title_required": True,
       "body_label": D_("Message"),
       "no_tabs": True,
   })


async def on_data_post(self, request):
    profile = self.getProfile(request)
    if profile is None:
        self.pageError(request, C.HTTP_FORBIDDEN)
    rdata = self.getRData(request)
    service = rdata["service"].full() if rdata["service"] else ""
    node = rdata["node"]
    title, body = self.getPostedData(request, ("title", "body"))
    title = title.strip()
    body = body.strip()
    if not title or not body:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    topic_data = {"title": title, "content": body}
    try:
        await self.host.bridgeCall(
            "forumTopicCreate", service, node, topic_data, profile
        )
    except Exception as e:
        if "forbidden" in str(e):
            self.pageError(request, 401)
        else:
            raise e

    rdata["post_redirect_page"] = (self.getPageByName("forum_topics"), service, node)



#!/usr/bin/env python3

from sat.core.i18n import _
from libervia.server.constants import Const as C
from twisted.words.protocols.jabber import jid
from twisted.internet import defer
from libervia.server import session_iface
from sat.tools.common import data_format
from sat.core.log import getLogger

log = getLogger(__name__)

name = "blog"
access = C.PAGES_ACCESS_PUBLIC
template = "blog/discover.html"


async def prepare_render(self, request):
    profile = self.getProfile(request)
    template_data = request.template_data
    if profile is not None:
        __, entities_own, entities_roster = await self.host.bridgeCall(
            "discoFindByFeatures",
            [],
            [("pubsub", "pep")],
            True,
            False,
            True,
            True,
            True,
            profile,
        )
        entities = template_data["disco_entities"] = (
            list(entities_own.keys()) + list(entities_roster.keys())
        )
        entities_url = template_data["entities_url"] = {}
        identities = self.host.getSessionData(
            request, session_iface.ISATSession
        ).identities
        d_list = {}
        for entity_jid_s in entities:
            entities_url[entity_jid_s] = self.getPageByName("blog_view").getURL(
                entity_jid_s
            )
            if entity_jid_s not in identities:
                d_list[entity_jid_s] = self.host.bridgeCall(
                        "identityGet",
                        entity_jid_s,
                        [],
                        True,
                        profile)
        identities_data = await defer.DeferredList(d_list.values())
        entities_idx = list(d_list.keys())
        for idx, (success, identity_raw) in enumerate(identities_data):
            entity_jid_s = entities_idx[idx]
            if not success:
                log.warning(_("Can't retrieve identity of {entity}")
                    .format(entity=entity_jid_s))
            else:
                identities[entity_jid_s] = data_format.deserialise(identity_raw)

        template_data["url_blog_edit"] = self.getSubPageURL(request, "blog_edit")


def on_data_post(self, request):
    jid_str = self.getPostedData(request, "jid")
    try:
        jid_ = jid.JID(jid_str)
    except RuntimeError:
        self.pageError(request, C.HTTP_BAD_REQUEST)
    url = self.getPageByName("blog_view").getURL(jid_.full())
    self.HTTPRedirect(request, url)



"""Integrate templating system using nunjucks"""

from js_modules.nunjucks import nunjucks
from browser import window, document
import javascript


safe = nunjucks.runtime.SafeString.new
env = nunjucks.configure(
    window.templates_root_url,
    {
        'autoescape': True,
        'trimBlocks': True,
        'lstripBlocks': True,
        'web': {'useCache': True},
    })

nunjucks.installJinjaCompat()
env.addGlobal("profile", window.profile)
env.addGlobal("csrf_token", window.csrf_token)
# FIXME: integrate gettext or equivalent here
env.addGlobal("_", lambda txt: txt)


class Indexer:
    """Index global to a page"""

    def __init__(self):
        self._indexes = {}

    def next(self, value):
        if value not in self._indexes:
            self._indexes[value] = 0
            return 0
        self._indexes[value] += 1
        return self._indexes[value]

    def current(self, value):
        return self._indexes.get(value)


gidx = Indexer()
# suffix use to avoid collision with IDs generated in static page
SCRIPT_SUFF = "__script__"

def escape_html(txt):
    return (
        txt
        .replace('&', '&amp;')
        .replace('<', '&lt;')
        .replace('>', '&gt;')
        .replace('"', '&quot;')
    )


def get_args(n_args, *sig_args, **sig_kwargs):
    """Retrieve function args when they are transmitted using nunjucks convention

    cf. https://mozilla.github.io/nunjucks/templating.html#keyword-arguments
    @param n_args: argument from nunjucks call
    @param sig_args: expected positional arguments
    @param sig_kwargs: expected keyword arguments
    @return: all expected arguments, with default value if not specified in nunjucks
    """
    # nunjucks set kwargs in last argument
    given_args = list(n_args)
    try:
        given_kwargs = given_args.pop().to_dict()
    except (AttributeError, IndexError):
        # we don't have a dict as last argument
        # that happens when there is no keyword argument
        given_args = list(n_args)
        given_kwargs = {}
    ret = given_args[:len(sig_args)]
    # we check if we have remaining positional arguments
    # in which case they may be specified in keyword arguments
    for name in sig_args[len(given_args):]:
        try:
            value = given_kwargs.pop(name)
        except KeyError:
            raise ValueError(f"missing positional arguments {name!r}")
        ret.append(value)

    extra_pos_args = given_args[len(sig_args):]
    # and now the keyword arguments
    for name, default in sig_kwargs.items():
        if extra_pos_args:
            # kw args has been specified with a positional argument
            ret.append(extra_pos_args.pop(0))
            continue
        value = given_kwargs.get(name, default)
        ret.append(value)

    return ret


def _next_gidx(value):
    """Use next current global index as suffix"""
    next_ = gidx.next(value)
    return f"{value}{SCRIPT_SUFF}" if next_ == 0 else f"{value}_{SCRIPT_SUFF}{next_}"

env.addFilter("next_gidx", _next_gidx)


def _cur_gidx(value):
    """Use current current global index as suffix"""
    current = gidx.current(value)
    return f"{value}{SCRIPT_SUFF}" if not current else f"{value}_{SCRIPT_SUFF}{current}"

env.addFilter("cur_gidx", _cur_gidx)


def _xmlattr(d, autospace=True):
    if not d:
        return
    d = d.to_dict()
    ret = [''] if autospace else []
    for key, value in d.items():
        if value is not None:
            ret.append(f'{escape_html(key)}="{escape_html(str(value))}"')

    return safe(' '.join(ret))

env.addFilter("xmlattr", _xmlattr)


def _tojson(value):
    return safe(escape_html(window.JSON.stringify(value)))

env.addFilter("tojson", _tojson)


def _icon_use(name, cls=""):
    kwargs = cls.to_dict()
    cls = kwargs.get('cls')
    return safe(
        '<svg class="svg-icon{cls}" xmlns="http://www.w3.org/2000/svg" '
        'viewBox="0 0 100 100">\n'
        '    <use href="#{name}"/>'
        '</svg>\n'.format(name=name, cls=(" " + cls) if cls else "")
    )

env.addGlobal("icon", _icon_use)


def _date_fmt(
    timestamp, *args
):
    """Date formatting

    cf. sat.tools.common.date_utils for arguments details
    """
    fmt, date_only, auto_limit, auto_old_fmt, auto_new_fmt = get_args(
        args, fmt="short", date_only=False, auto_limit=7, auto_old_fmt="short",
        auto_new_fmt="relative",
    )
    from js_modules.moment import moment
    date = moment.unix(timestamp)

    if fmt == "auto_day":
        fmt, auto_limit, auto_old_fmt, auto_new_fmt = "auto", 0, "short", "HH:mm"
    if fmt == "auto":
        limit = moment().startOf('day').subtract(auto_limit, 'days')
        m_fmt = auto_old_fmt if date < limit else auto_new_fmt

    if fmt == "short":
        m_fmt = "DD/MM/YY" if date_only else "DD/MM/YY HH:mm"
    elif fmt == "medium":
        m_fmt = "ll" if date_only else "lll"
    elif fmt == "long":
        m_fmt = "LL" if date_only else "LLL"
    elif fmt == "full":
        m_fmt = "dddd, LL" if date_only else "LLLL"
    elif fmt == "relative":
        return date.fromNow()
    elif fmt == "iso":
        if date_only:
            m_fmt == "YYYY-MM-DD"
        else:
            return date.toISOString()
    else:
        raise NotImplementedError("free format is not implemented yet")

    return date.format(m_fmt)

env.addFilter("date_fmt", _date_fmt)


class I18nExtension:
    """Extension to handle the {% trans %}{% endtrans %} statement"""
    # FIXME: for now there is no translation, this extension only returns the string
    #        unmodified
    tags = ['trans']

    def parse(self, parser, nodes, lexer):
        tok = parser.nextToken()
        args = parser.parseSignature(None, True)
        parser.advanceAfterBlockEnd(tok.value)
        body = parser.parseUntilBlocks('endtrans')
        parser.advanceAfterBlockEnd()
        return nodes.CallExtension.new(self._js_ext, 'run', args, [body])

    def run(self, context, *args):
        body = args[-1]
        return body()

    @classmethod
    def install(cls, env):
        ext = cls()
        ext_dict = {
            "tags": ext.tags,
            "parse": ext.parse,
            "run": ext.run
        }
        ext._js_ext = javascript.pyobj2jsobj(ext_dict)
        env.addExtension(cls.__name__, ext._js_ext)

I18nExtension.install(env)


class Template:

    def __init__(self, tpl_name):
        self._tpl = env.getTemplate(tpl_name, True)

    def render(self, context):
        return self._tpl.render(context)

    def get_elt(self, context=None):
        if context is None:
            context = {}
        raw_html = self.render(context)
        template_elt = document.createElement('template')
        template_elt.innerHTML = raw_html
        return template_elt.content.firstElementChild



#!/usr/bin/env python3

# Libervia: a SàT frontend
# Copyright (C) 2009-2021 Jérôme Poisson (goffi@goffi.org)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..common import constants


class Const(constants.Const):

    APP_NAME = "Libervia Web"
    APP_COMPONENT = "web"
    APP_NAME_ALT = APP_NAME
    APP_NAME_FILE = "libervia_web"
    CONFIG_SECTION = APP_COMPONENT.lower()
    SERVICE_PROFILE = "libervia"  # the SàT profile that is used for exporting the service

    SESSION_TIMEOUT = 7200  # Session's timeout, after that the user will be disconnected
    HTML_DIR = "html/"
    THEMES_DIR = "themes/"
    THEMES_URL = "themes"
    MEDIA_DIR = "media/"
    CARDS_DIR = "games/cards/tarot"
    PAGES_DIR = "pages"
    TASKS_DIR = "tasks"
    LIBERVIA_CACHE = "libervia"
    SITE_NAME_DEFAULT = "default"
    # generated files will be accessible there
    BUILD_DIR = "__b"
    BUILD_DIR_DYN = "dyn"
    # directory where build files are served to the client
    PRODUCTION_BUILD_DIR = "sites"
    # directory used for files needed temporarily (e.g. for compiling other files)
    DEV_BUILD_DIR = "dev_build"

    TPL_RESOURCE = '_t'

    ERRNUM_BRIDGE_ERRBACK = 0  # FIXME
    ERRNUM_LIBERVIA = 0  # FIXME

    # Security limit for Libervia (get/set params)
    SECURITY_LIMIT = 5

    # Security limit for Libervia server_side
    SERVER_SECURITY_LIMIT = constants.Const.NO_SECURITY_LIMIT

    # keys for cache values we can get from browser
    ALLOWED_ENTITY_DATA = {"avatar", "nick"}

    STATIC_RSM_MAX_LIMIT = 100
    STATIC_RSM_MAX_DEFAULT = 10
    STATIC_RSM_MAX_COMMENTS_DEFAULT = 10

    ## Libervia pages ##
    PAGES_META_FILE = "page_meta.py"
    PAGES_BROWSER_DIR = "_browser"
    PAGES_BROWSER_META_FILE = "browser_meta.json"
    PAGES_ACCESS_NONE = (
        "none"
    )  #  no access to this page (using its path will return a 404 error)
    PAGES_ACCESS_PUBLIC = "public"
    PAGES_ACCESS_PROFILE = (
        "profile"
    )  # a session with an existing profile must be started
    PAGES_ACCESS_ADMIN = "admin"  #  only profiles set in admins_list can access the page
    PAGES_ACCESS_ALL = (
        PAGES_ACCESS_NONE,
        PAGES_ACCESS_PUBLIC,
        PAGES_ACCESS_PROFILE,
        PAGES_ACCESS_ADMIN,
    )
    # names of the page to use for menu
    DEFAULT_MENU = [
        "login",
        "chat",
        "blog",
        "forums",
        "photos",
        "files",
        "events",
        "lists",
        "merge-requests",
        # XXX: app is not available anymore since removal of pyjamas code with Python 3
        #    port. It should come back at a later point with an alternative (Brython
        #    probably).
    ]

    ## Session flags ##
    FLAG_CONFIRM = "CONFIRM"

    ## Data post ##
    POST_NO_CONFIRM = "POST_NO_CONFIRM"

    ## HTTP methods ##
    HTTP_METHOD_GET = b"GET"
    HTTP_METHOD_POST = b"POST"

    ## HTTP codes ##
    HTTP_SEE_OTHER = 303
    HTTP_NOT_MODIFIED = 304
    HTTP_BAD_REQUEST = 400
    HTTP_UNAUTHORIZED = 401
    HTTP_FORBIDDEN = 403
    HTTP_NOT_FOUND = 404
    HTTP_INTERNAL_ERROR = 500
    HTTP_PROXY_ERROR = 502
    HTTP_SERVICE_UNAVAILABLE = 503

    ## HTTP HEADERS ##
    H_FORWARDED = "Forwarded"
    H_X_FORWARDED_FOR = "X-Forwarded-For"
    H_X_FORWARDED_HOST = "X-Forwarded-Host"
    H_X_FORWARDED_PROTO = "X-Forwarded-Proto"


    ## Cache ##
    CACHE_PUBSUB = 0

    ## Date/Time ##
    HTTP_DAYS = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
    HTTP_MONTH = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct",
                  "Nov", "Dec")



#!/usr/bin/env python3


from sat.core.i18n import _
from sat.core import exceptions
from libervia.server.constants import Const as C
from libervia.server import session_iface
from twisted.internet import defer
from sat.core.log import getLogger

log = getLogger(__name__)

"""SàT log-in page, with link to create an account"""

name = "login"
access = C.PAGES_ACCESS_PUBLIC
template = "login/login.html"


def prepare_render(self, request):
    template_data = request.template_data

    #  we redirect to logged page if a session is active
    profile = self.getProfile(request)
    if profile is not None:
        self.pageRedirect("/login/logged", request)

    # login error message
    session_data = self.host.getSessionData(request, session_iface.ISATSession)
    login_error = session_data.popPageData(self, "login_error")
    if login_error is not None:
        template_data["S_C"] = C  # we need server constants in template
        template_data["login_error"] = login_error
    template_data["empty_password_allowed"] = bool(
        self.host.options["empty_password_allowed_warning_dangerous_list"]
    )

    # register page url
    if self.host.options["allow_registration"]:
        template_data["register_url"] = self.getPageRedirectURL(request, "register")

    #  if login is set, we put it in template to prefill field
    template_data["login"] = session_data.popPageData(self, "login")


def login_error(self, request, error_const):
    """set login_error in page data

    @param error_const(unicode): one of login error constant
    @return C.POST_NO_CONFIRM: avoid confirm message
    """
    session_data = self.host.getSessionData(request, session_iface.ISATSession)
    session_data.setPageData(self, "login_error", error_const)
    return C.POST_NO_CONFIRM


@defer.inlineCallbacks
def on_data_post(self, request):
    profile = self.getProfile(request)
    type_ = self.getPostedData(request, "type")
    if type_ == "disconnect":
        if profile is None:
            log.warning(_("Disconnect called when no profile is logged"))
            self.pageError(request, C.HTTP_BAD_REQUEST)
        else:
            self.host.purgeSession(request)
            defer.returnValue(C.POST_NO_CONFIRM)
    elif type_ == "login":
        login, password = self.getPostedData(request, ("login", "password"))
        try:
            status = yield self.host.connect(request, login, password)
        except exceptions.ProfileUnknownError:
            # the profile doesn't exist, we return the same error as for invalid password
            # to avoid bruteforcing valid profiles
            log.warning(f"login tentative with invalid profile: {login!r}")
            defer.returnValue(login_error(self, request, C.PROFILE_AUTH_ERROR))
        except ValueError as e:
            message = str(e)
            if message in (C.XMPP_AUTH_ERROR, C.PROFILE_AUTH_ERROR):
                defer.returnValue(login_error(self, request, message))
            else:
                # this error was not expected!
                raise e
        except exceptions.TimeOutError:
            defer.returnValue(login_error(self, request, C.NO_REPLY))
        else:
            if status in (C.PROFILE_LOGGED, C.PROFILE_LOGGED_EXT_JID, C.SESSION_ACTIVE):
                # Profile has been logged correctly
                self.redirectOrContinue(request)
            else:
                log.error(_("Unhandled status: {status}".format(status=status)))
    else:
        self.pageError(request, C.HTTP_BAD_REQUEST)



#!/usr/bin/env python3


from libervia.server.constants import Const as C
from twisted.internet import defer
from sat.core.log import getLogger
from sat.tools.common import date_utils

"""creation of new events"""

name = "event_new"
access = C.PAGES_ACCESS_PROFILE
template = "event/create.html"
log = getLogger(__name__)


@defer.inlineCallbacks
def on_data_post(self, request):
    request_data = self.getRData(request)
    profile = self.getProfile(request)
    title, location, body, date, main_img, bg_img = self.getPostedData(
        request, ("name", "location", "body", "date", "main_image", "bg_image")
    )
    timestamp = date_utils.date_parse(date)
    data = {"name": title, "description": body, "location": location}

    for value, var in ((main_img, "image"), (bg_img, "background-image")):
        value = value.strip()
        if not value:
            continue
        if not value.startswith("http"):
            self.pageError(request, C.HTTP_BAD_REQUEST)
        data[var] = value
    data["register"] = C.BOOL_TRUE
    node = yield self.host.bridgeCall("eventCreate", timestamp, data, "", "", "", profile)
    log.info("Event node created at {node}".format(node=node))

    request_data["post_redirect_page"] = (self.getPageByName("event_admin"), "@", node)
    defer.returnValue(C.POST_NO_CONFIRM)



from browser import DOMNode, document, aio
from javascript import JSON
from aio_bridge import Bridge, BridgeException
import dialog

bridge = Bridge()


async def on_delete(evt):
    evt.stopPropagation()
    evt.preventDefault()
    target = evt.currentTarget
    item_elt = DOMNode(target.closest('.item'))
    item_elt.classList.add("selected_for_deletion")
    item = JSON.parse(item_elt.dataset.item)
    confirmed = await dialog.Confirm(
        f"List {item['name']!r} will be deleted, are you sure?",
        ok_label="delete",
    ).ashow()

    if not confirmed:
        item_elt.classList.remove("selected_for_deletion")
        return

    try:
        await bridge.interestRetract("", item['id'])
    except BridgeException as e:
        dialog.notification.show(
            f"Can't remove list {item['name']!r} from personal interests: {e}",
            "error"
        )
    else:
        print(f"{item['name']!r} removed successfuly from list of interests")
        item_elt.classList.add("state_deleted")
        item_elt.bind("transitionend", lambda evt: item_elt.remove())
        if item.get("creator", False):
            try:
                await bridge.psNodeDelete(
                    item['service'],
                    item['node'],
                )
            except BridgeException as e:
                dialog.notification.show(
                    f"Error while deleting {item['name']!r}: {e}",
                    "error"
                )
            else:
                dialog.notification.show(f"{item['name']!r} has been deleted")


for elt in document.select('.action_delete'):
    elt.bind("click", lambda evt: aio.run(on_delete(evt)))
